#SkIndigo Ruby code is copyright Dale Martens 2011, all rights reserved, and
#Glare Technologies Limited, 2014 - 
#-----------------------------------------------------------------------------

#######################Indigo Exporter Class#####################


class IndigoInvalidTabulatedException < RuntimeError
end



class IndigoExporter

	attr_reader :preview_material, :indigo_version, :path, :exported_mats, :model, :irs, :start, :end, :tex_path, :nkdata, :textures, :ies, :meshbuilder, :physics, :clay_render, :light_source, :preview_scene
	attr_writer :tex_path, :path, :preview_material, :start, :end, :meshbuilder, :physics, :clay_render, :preview_scene

	SKY_LAYER_INDEX=1
	SUN_LAYER_INDEX=0
	BACKGROUND_LAYER_INDEX=2

	def initialize(path)
		@path = path	# The path of the file that the Exporter will write to
		@tex_path = ""	# The directory that the Exporter will write textures to
		@nkdata = []
		@textures = []
		@ies = []
		@null_exported = false
		@preview_material = nil
		@indigo_version = SkIndigo::VERSION
		@model = Sketchup.active_model
		@materials = @model.materials
		@irs = IndigoRenderSettings.new
		@scatter_settings = IndigoScatterSettings.new
		@used_layers_set = Set.new  # Keeps track of light layers that are actually used
		@used_layers_array = []

		@scale = 0.0254 * Indigo.scale
		@igm = false
		@frame = 0  # The current animation export frame
		@meshbuilder = nil
		@physics = nil
		@clay_render = false
		@light_source = false
		@preview_scene = false
		@export_warnings = [] # List of strings
		
		@material_uids = {} # Map from [SketchUp material, generate_uvs] to exported UID
		@next_uid = 100
		@default_mat_uid = -1
		
		# Used for checking blend material children during material export
		@all_mat_names = Sketchup.active_model.materials.collect {|m| m.name}
		
		@last_model_uid_for_defn = {}
	end
	
	
	def export_warnings
		return @export_warnings
	end

	
	def get_and_incr_next_uid
		uid = @next_uid
		@next_uid = @next_uid + 1
		return uid
	end


	###Always wraps in a CDATA string, to preserve whitespace, XML special characters etc..
	def write_name_string(out, depth, name, string)
		
		out.print("\t" * depth) # Write the tabs
		out.print "<#{name}><![CDATA[#{string}]]></#{name}>\n"
	end


	def validate_light_source()

		if @light_source==false
			UI.messagebox("Scene must contain a light source.  Export aborted.")
			raise
		end
	end


	def settings
		return @irs
	end


	#############################################Material Exporter######################
	# mat - type IndigoMaterial
	def export_material(mat, out, generate_uvs=false)

		if mat.type == "Blend"   # Check if A material and B material exist and export them first
			write_null_mat(out) if ((mat.a_name=="None") || (mat.b_name=="None"))
			a_mat = IndigoMaterial.find(mat.a_name)
			b_mat = IndigoMaterial.find(mat.b_name)
			
			write_null_mat(out) if (a_mat.nil? || b_mat.nil?)
			
			export_material(a_mat, out, generate_uvs) if a_mat
			export_material(b_mat, out, generate_uvs) if b_mat
		end
		
		if mat.type == "Coating"  # Export the substrate first
			substrate = IndigoMaterial.find(mat.substrate)
			export_material(substrate, out, generate_uvs) if substrate
		end
		
		if mat.type == "Double-Sided Thin"  # Export the front and back materials first
			front_mat = IndigoMaterial.find(mat.get("front_material_name"))
			back_mat  = IndigoMaterial.find(mat.get("back_material_name"))
			export_material(front_mat, out, generate_uvs) if front_mat
			export_material(back_mat, out, generate_uvs) if back_mat
		end
		
		if !@material_uids.key?([mat.mat, generate_uvs]) # If not already exported
			write_mat(mat, out, generate_uvs) # Export the material	
		end
	end
	
	
	def write_clay_mat(out, mat_uid)
		out.puts "	<material>"
		out.puts "		<uid>#{mat_uid}</uid>"
		out.puts "		<name>Clay</name>"
		out.puts "		<diffuse>"
		out.puts "			<albedo>"
		out.puts "				<constant>"
		out.puts "					<uniform><value>0.5</value></uniform>"
		out.puts "				</constant>"
		out.puts "			</albedo>"
		out.puts "		</diffuse>"
		out.puts "	</material>"
	end
	
	
	# Rewrite paths in an XML DOM to be relative to the directory 'igm_dir'.
	def rewrite_paths(elem, igm_dir)
		elem.children.each do |child|
			
			# Rewrite all 'path' elements
			path_elems = child.find_all("path")
			path_elems.each do |path_elem|
				#puts "found path_elem.  path_elem.element_text: #{path_elem.element_text}"
				newpath = File.join(igm_dir, path_elem.element_text)
				#puts "newpath: #{newpath}"
				path_elem.element_text = newpath
			end
			
			# Recurse
			rewrite_paths(child, igm_dir)
		end
	end
	
	
	# constant_rgb - From mat.rgb, Sketchup color
	# tex_index - Index of exported texture, or nil if no texture
	# shader - from something like mat.albedo_texture.shader_code
	# mat_texture - something like mat.albedo_texture.  Needed?
	def export_wavelen_dep_param(elem_name, constant_rgb, tex_index, shader, mat_texture, out)
		out.puts "			<#{elem_name}>"
		if tex_index
			export_texture_index(tex_index, out)
		elsif shader
			export_shader(shader, out)
		else
			export_constant_wavelen_dep_param(constant_rgb, mat_texture, out)
		end
		out.puts "			</#{elem_name}>"
	end
	
	
	# tex_index - Index of exported texture, or nil if no texture
	# shader - from something like mat.albedo_texture.shader_code
	# mat_map - something like mat.albedo_texture
	def export_wavelen_indep_param(elem_name, map_index, shader, mat_map, out)
		out.puts "			<#{elem_name}>"
		if map_index
			export_texture_index(map_index, out)
		elsif shader
			export_shader(shader, out)
		else
			out.puts "				<constant>#{mat_map.b}</constant>"
		end
		out.puts "			</#{elem_name}>"
	end
	
	def export_base_emission(sketchup_mat_rgb, mat_base_emission, out)
		out.puts "			<base_emission>"
		export_constant_wavelen_dep_param(sketchup_mat_rgb, mat_base_emission, out)
		out.puts "			</base_emission>"
	end
	
	
	#####################Material Writer############################
	# mat - type IndigoMaterial
	def write_mat(mat, out, generate_uvs)
	
		begin 
			# Export all materials as clay, apart from emitting materials and specular materials.
			if self.clay_render && !mat.emitter? && (mat.type != "Specular")
				mat_uid = get_and_incr_next_uid()
				write_clay_mat(out, mat_uid)
				@material_uids[[mat.mat, generate_uvs]] = mat_uid
				return
			end

			@tex_index = 0
			export_name = (mat == preview_material) ? "previewmaterial" : mat.name
			
			sketchup_mat_rgb = mat.rgb # Get the colour of the SketchUp material
			opacity = mat.alpha
			type = mat.type

			albedo_tex = nil 			
			bump_map = nil
			exponent_map = nil
			normal_map = nil
			sigma_map = nil
			disp_map = nil
			emission_map = nil
			blend_map = nil
			specular_map = nil
			abs_layer_trans_map = nil
			roughness_map = nil
			fresnel_map = nil
			thickness_map = nil
			transmittance_map = nil
			r_f_map = nil
			front_fresnel_scale_map = nil
			back_fresnel_scale_map = nil
			front_roughness_map = nil
			back_roughness_map = nil

			albedo_shader = nil
			bump_shader = nil
			normal_shader = nil
			exponent_shader = nil
			sigma_shader = nil
			disp_shader = nil
			blend_shader = nil
			emission_shader = nil
			specular_shader = nil
			abs_layer_trans_shader = nil
			abs_shader = nil
			sss_shader = nil
			roughness_shader = nil
			fresnel_shader = nil
			thickness_shader = nil
			transmittance_shader = nil
			r_f_shader = nil
			front_fresnel_scale_shader = nil
			back_fresnel_scale_shader = nil
			front_roughness_shader = nil
			back_roughness_shader = nil
			
			if (type=="Specular") or (type=="Glossy Transparent") or (type == "Fast SSS")
				#alpha_factor=opacity*opacity*1000
				#sss_spectrum=mat.sss_spectrum.to_s.split(' ')
				#if sss_spectrum[0] and sss_spectrum[1] and sss_spectrum[2]
					#rgb_spectrum=true
				#else
					#rgb_spectrum=false
				#end
				#g_spectrum=mat.g_spectrum.to_s.split(' ')
				#if g_spectrum[0] and g_spectrum[1] and g_spectrum[2]
				#	g_rgb_spectrum=true
				#else
				#	g_rgb_spectrum=false
				#end

				# Export the medium
				medium_uid = get_and_incr_next_uid()
				
				medium_name = export_name + " medium"
				out.puts "	<medium>"
				out.puts "		<uid>#{medium_uid}</uid>"
				write_name_string(out, 2, "name", medium_name)
				out.puts "		<precedence>#{mat.precedence}</precedence>"
				out.puts "		<basic>"
				out.puts "			<ior>#{mat.ior}</ior>"
				out.puts "			<cauchy_b_coeff>#{mat.cauchy_b_coeff}</cauchy_b_coeff>"
				out.puts "			<fast_sss>true</fast_sss>" if (type == "Fast SSS")
				export_absorption(mat, out)
				if mat.sss?
					out.puts "			<subsurface_scattering>"
					export_sss(mat, out)
					out.puts "				<phase_function>"
					if mat.phase_function == "Uniform"
						out.puts "					<uniform/>"
					else
						out.puts "					<henyey_greenstein>"
						out.puts "						<g_spectrum>"
						export_constant_wdp(mat.g_spectrum_map, out, false)
						out.puts "						</g_spectrum>"
						out.puts "					</henyey_greenstein>"
					end

					out.puts "				</phase_function>"
					out.puts "			</subsurface_scattering>"
				end
				out.puts "		</basic>"
				out.puts "	</medium>"
				out.puts ""
		
			end # Finished exporting the medium

			# Material Export Starts Here
		
			case type
			when "Linked IGM"
				
				ligm = mat.linked_igm
				original_ligm_path = ligm
				
				if !FileTest.exist?(ligm)
					# Linked IGM was not found (maybe files were moved on disk?)  Try and fix up path:
					ligm = SkIndigo.find_file_in_dir_tree(
						File.dirname(Sketchup.active_model.path), # Search directory
						File.basename(ligm) # Target file name
					)
				end
				
				
				if ligm.nil? || !FileTest.exist?(ligm)
					mat_uid = get_and_incr_next_uid()
					write_clay_mat(out, mat_uid)
					@material_uids[[mat.mat, generate_uvs]] = mat_uid
					@export_warnings << "Could not find linked igm material '#{original_ligm_path}' for material '#{mat.display_name}'"
					return
				end
				
				# Read the linked IGM data, and write into the current output scene file, adding some UID tags and rewriting texture paths.
				ligm_file = File.new(ligm, "r")
				data = ligm_file.read
				ligm_file.close

				data = SkIndigo.remove_invalid_utf8_sequences(data)
				
				xml_parser = DMxml::Parser.new(data)
				doc = xml_parser.parse_document()
		
				# Texture paths would have been relative to IGM directory.
				# Rewrite texture paths to be relative to the export dir.
				igm_dir = File.dirname(ligm)
				
				rewrite_paths(doc.root, igm_dir) # Rewrite paths in the XML DOM to be relative to the directory 'igm_dir'.
				
				# Write out the XML
				out.puts "	<!--================== linked IGM data ==================-->"
				doc.root.children.each do |elem|
					mat_uid = get_and_incr_next_uid() # Generate a new UID for this node (medium or material)
					
					# Add a UID tag to the element
					new_elem = DMxml::Element.new("uid")
					new_elem.element_text = mat_uid.to_s
					elem.children << new_elem
					
					out.puts elem.to_s(1) # Serialise (with depth=1)
				end
				out.puts "	<!--================== End linked IGM data ==================-->"
				
				# Remember the last material UID exported as the UID for this material.
				@material_uids[[mat.mat, generate_uvs]] = mat_uid
				
			when "Simple Thin"
			
				export_simple_thin_mat(mat, out, export_name, generate_uvs)
				
			when "Diffuse", "Specular", "Phong", "Metals", "Oren-Nayar", "Glossy Transparent", "Blend", "Coating", "Double-Sided Thin", "Fast SSS"
			
				# Export blended diffuse transmitters
				if (type == "Diffuse" or type == "Phong") && ((opacity < 1.0) and (opacity > 0.0))
					export_blended_diffuse_transmitter(mat, export_name, out, generate_uvs)
					export_clip_map(mat, out, export_name, generate_uvs)  if mat.use_clip? #if mat.clip_map? or mat.clip_shader? #exports the clip map
					return
				end
				
				mat_uid = get_and_incr_next_uid()
				
				out.puts "	<material>"
				out.puts "		<uid>#{mat_uid}</uid>"
				write_name_string(out, 2, "name", export_name)
				indigo_type = mat.indigo_type
				out.puts "		<#{indigo_type}>"
				unless type == "Blend"  #what about coating?  - can you have emitting coating?
					if mat.emitter?
						out.puts "			<layer>#{@used_layers_array.index(mat.layer.to_i)}</layer>" if !@igm
						#@used_layers.insert(mat.layer)
					end
				end
				if type == "Specular" or type == "Glossy Transparent" or type == "Fast SSS"
					if type == "Specular"
						out.print " 		<transparent>true</transparent>\n" if opacity < 1.0
						out.print "			<transparent>false</transparent>\n" if opacity == 1.0
						out.puts "			<arch_glass>#{mat.arch_glass?}</arch_glass>"
					end
					write_name_string(out, 3, "internal_medium_name", medium_name)
				end

				# Export all of the texture xml and collect the indices
				albedo_tex 				= export_texture(mat, mat.albedo_texture, out, generate_uvs) if mat.albedo_texture?
				specular_map 			= export_texture(mat, mat.specular_map, out, generate_uvs) if mat.specular_map?
				blend_map 				= export_texture(mat, mat.blend_map, out, generate_uvs) if mat.blend_map? 			
				bump_map 				= export_texture(mat, mat.bump_map, out, generate_uvs)  if mat.bump_map?
				normal_map 				= export_texture(mat, mat.normal_map, out, generate_uvs)  if mat.normal_map?
				exponent_map 			= export_texture(mat, mat.exponent_map, out, generate_uvs) if mat.exponent_map?
				sigma_map 				= export_texture(mat, mat.sigma_map, out, generate_uvs) if mat.sigma_map?
				disp_map 				= export_texture(mat, mat.disp_map, out, generate_uvs)  if mat.disp_map? #and (!mat.clip_map?) #don't export the displacement map if this material has a clip map,  export it with the blend material
				emission_map 			= export_texture(mat, mat.emission_map, out, generate_uvs) if mat.emission_map?
				abs_layer_trans_map 	= export_texture(mat, mat.abs_layer_trans_map, out, generate_uvs) if mat.abs_layer_trans_map?
				roughness_map 			= export_texture(mat, mat.roughness_map, out, generate_uvs) if mat.roughness_map?
				fresnel_map 			= export_texture(mat, mat.fresnel_map, out, generate_uvs) if mat.fresnel_map?
				thickness_map 			= export_texture(mat, mat.thickness_map, out, generate_uvs) if mat.thickness_map?
				transmittance_map 		= export_texture(mat, mat.transmittance_map, out, generate_uvs) if mat.transmittance_map?
				r_f_map 				= export_texture(mat, mat.r_f_map, out, generate_uvs) if mat.r_f_map?
				front_fresnel_scale_map = export_texture(mat, mat.front_fresnel_scale_map, out, generate_uvs) if mat.front_fresnel_scale_map?
				back_fresnel_scale_map 	= export_texture(mat, mat.back_fresnel_scale_map, out, generate_uvs) if mat.back_fresnel_scale_map?
				front_roughness_map 	= export_texture(mat, mat.front_roughness_map, out, generate_uvs) if mat.front_roughness_map?
				back_roughness_map 		= export_texture(mat, mat.back_roughness_map, out, generate_uvs) if mat.back_roughness_map?
				
				# Collect all of the shaders
				albedo_shader 			= mat.albedo_texture.shader_code if mat.albedo_shader?
				specular_shader 		= mat.specular_map.shader_code if mat.specular_shader?
				blend_shader 			= mat.blend_map.shader_code if mat.blend_shader?
				bump_shader 			= mat.bump_map.shader_code if mat.bump_shader?
				normal_shader 			= mat.normal_map.shader_code if mat.normal_shader?
				exponent_shader 		= mat.exponent_map.shader_code if mat.exponent_shader?
				sigma_shader 			= mat.sigma_map.shader_code if mat.sigma_shader?
				disp_shader 			= mat.disp_map.shader_code if mat.disp_shader?
				emission_shader 		= mat.emission_map.shader_code if mat.emission_shader?
				abs_layer_trans_shader 	= mat.abs_layer_trans_map.shader_code if mat.abs_layer_trans_shader?
				roughness_shader 		= mat.roughness_map.shader_code if mat.roughness_shader?
				fresnel_shader 			= mat.fresnel_map.shader_code if mat.fresnel_shader?
				thickness_shader 		= mat.thickness_map.shader_code if mat.thickness_shader?
				transmittance_shader 	= mat.transmittance_map.shader_code if mat.transmittance_shader?
				r_f_shader 				= mat.r_f_map.shader_code if mat.r_f_shader?
				front_fresnel_scale_shader = mat.front_fresnel_scale_map.shader_code if mat.front_fresnel_scale_shader?
				back_fresnel_scale_shader = mat.back_fresnel_scale_map.shader_code if mat.back_fresnel_scale_shader?
				front_roughness_shader 	= mat.front_roughness_map.shader_code if mat.front_roughness_shader?
				back_roughness_shader 	= mat.back_roughness_map.shader_code if mat.back_roughness_shader?

				
			
				if type=="Blend"
					matA = mat.a_name
					matB = mat.b_name
					indexA = @all_mat_names.index(matA)
					indexB = @all_mat_names.index(matB)
					if (indexA==nil and matA!="None")
						UI.messagebox("Warning: Blend material #{mat.name} tried to use non-existent material named #{matA}.  The material #{matA} may have been deleted.")
						matA = "None" # Just export using Null material instead of the missing material.
					end
					if (indexB==nil and matB!="None")
						UI.messagebox("Warning: Blend material #{mat.name} tried to use non-existent material named #{matB}.  The material #{matB} may have been deleted.")
						matB = "None" # Just export using Null material instead of the missing material.
					end
					
					#NOTE: although we could export using UIDs, can just continue to use the name-based code.
					#skup_mat_a = IndigoMaterial.find(matA)
					#skup_mat_b = IndigoMaterial.find(matB)
					
					#a_mat_uid = @material_uids[skup_mat_a.mat]
					#b_mat_uid = @material_uids[skup_mat_b.mat]
					
					#out.print "				<a_mat_uid>#{a_mat_uid}</a_mat_uid>\n"
					#out.print "				<b_mat_uid>#{b_mat_uid}</b_mat_uid>\n"
					write_name_string(out, 4, "a_name", matA)	
					write_name_string(out, 4, "b_name", matB)			
					export_wavelen_indep_param("blend", blend_map, blend_shader, mat.blend_map, out)
					out.puts "				<step_blend>#{mat.step_blend?}</step_blend>"
				end

				# Export albedo param
				if type == "Diffuse" or type == "Oren-Nayar" or type == "Phong" or type == "Fast SSS"
					elem_name = (type == "Phong") ? "diffuse_albedo" : "albedo"
					export_wavelen_dep_param(elem_name, sketchup_mat_rgb, albedo_tex, albedo_shader, mat.albedo_texture, out)
				end

				# Export IOR param
				if type == "Phong" or type == "Coating" or type == "Double-Sided Thin"
					out.puts "			<ior>#{mat.ior}</ior>"
				end
				
				if type == "Metals"
					nkdata = mat.nk_data    #returns an NKdata object
					if mat.use_nkdata?
						out.puts "				<nk_data>nkdata\/#{nkdata.filename}</nk_data>"
						@nkdata.push(nkdata.path) if (not @nkdata.index(nkdata.path))    #copies stores the nkdata path for future export
					else
						export_wavelen_dep_param("specular_reflectivity", sketchup_mat_rgb, specular_map, specular_shader, mat.specular_map, out)
					end
				end 
				if type=="Oren-Nayar"
					export_wavelen_indep_param("sigma", sigma_map, sigma_shader, mat.sigma_map, out)
				end
				if type=="Phong" or type=="Glossy Transparent" or type=="Metals"
					#export_wavelen_indep_param("exponent", exponent_map, exponent_shader, mat.exponent_map, out)
					#puts "mat: #{mat.inspect}"
					
					# If this material has a roughness map, export it
					if mat.roughness_map?
						#puts "mat has roughness map."
						export_wavelen_indep_param("roughness", roughness_map, roughness_shader, mat.roughness_map, out)
					else
						#puts "mat does not have roughness map."
						if mat.exponent_map?
							#puts "mat has exponent map."
							export_wavelen_indep_param("exponent", exponent_map, exponent_shader, mat.exponent_map, out)
						else
							#puts "mat does not have exponent map."
							#puts "mat.is_roughness_param_stored?: #{mat.is_roughness_param_stored?}"
							if mat.is_roughness_param_stored?
								#puts "exporting roughesss."
								export_wavelen_indep_param("roughness", roughness_map, roughness_shader, mat.roughness_map, out)
							else
								#puts "exporting exponent."
								export_wavelen_indep_param("exponent", exponent_map, exponent_shader, mat.exponent_map, out)
							end
						end
					end
				end
				
				if type == "Fast SSS"
					export_wavelen_indep_param("roughness", roughness_map, roughness_shader, mat.roughness_map, out)
				end
				
				if type=="Glossy Transparent" or type=="Specular"
					#TODO - validate absorption layer parameter
					export_wavelen_dep_param("absorption_layer_transmittance", sketchup_mat_rgb, abs_layer_trans_map, abs_layer_trans_shader, mat.abs_layer_trans_map, out) if mat.use_abs_layer_trans?
				end
				
				# Export fresnel_scale
				if type=="Coating" or type=="Phong" or type == "Fast SSS"
					export_wavelen_indep_param("fresnel_scale", fresnel_map, fresnel_shader, mat.fresnel_map, out)
				end
				
				if type=="Coating"
					export_coating_absorption(mat,out)
					export_wavelen_indep_param("thickness", thickness_map, thickness_shader, mat.thickness_map, out)
					export_wavelen_indep_param("roughness", roughness_map, roughness_shader, mat.roughness_map, out)
					out.puts "				<interference>#{mat.interference?}</interference>"
					
					if(mat.substrate == "None") # This is the case when the user has not set a substate material properly.
						# In this case just use the default (clay) material as the substate, to avoid rendering errors.
						write_name_string(out, 3, "substrate_material_name", "Default")
					else
						write_name_string(out, 3, "substrate_material_name", mat.substrate)
					end
					
				elsif type=="Double-Sided Thin"
					write_name_string(out, 3, "front_material_name", mat.get("front_material_name"))
					write_name_string(out, 3, "back_material_name",  mat.get("back_material_name"))
					export_wavelen_indep_param("r_f", r_f_map, r_f_shader, mat.r_f_map, out)
					export_wavelen_dep_param("transmittance", sketchup_mat_rgb, transmittance_map, transmittance_shader, mat.transmittance_map, out)
					export_wavelen_indep_param("front_roughness", front_roughness_map, front_roughness_shader, mat.front_roughness_map, out)
					export_wavelen_indep_param("back_roughness", back_roughness_map, back_roughness_shader, mat.back_roughness_map, out)
					export_wavelen_indep_param("front_fresnel_scale", front_fresnel_scale_map, front_fresnel_scale_shader, mat.front_fresnel_scale_map, out)
					export_wavelen_indep_param("back_fresnel_scale", back_fresnel_scale_map, back_fresnel_scale_shader, mat.back_fresnel_scale_map, out)
				end
				
				# Export the common parameters
				export_wavelen_indep_param("bump", bump_map, bump_shader, mat.bump_map, out) if mat.use_bump?
				export_wavelen_indep_param("normal_map", normal_map, normal_shader, mat.normal_map, out) if mat.use_normal? 
				export_wavelen_indep_param("displacement", disp_map, disp_shader, mat.disp_map, out) if mat.use_displace?
				export_base_emission(sketchup_mat_rgb, mat.base_emission, out) if mat.emitter?
				export_wavelen_dep_param("emission", sketchup_mat_rgb, emission_map, emission_shader, mat.emission_map, out) if mat.use_emission?
					
				out.puts "		</#{indigo_type}>"
				
				# Export some emission stuff
				if(mat.emission_sampling_factor != 1.0)
					out.puts "		<emission_sampling_factor>#{mat.emission_sampling_factor}</emission_sampling_factor>"
				end
				if(mat.backface_emit?)
					out.puts "		<backface_emit>#{mat.backface_emit?}</backface_emit>"
				end
				
				out.puts "	</material>"
				out.puts ""
				
				@material_uids[[mat.mat, generate_uvs]] = mat_uid
				
				export_clip_map(mat, out, export_name, generate_uvs) if mat.use_clip?
				
			when "Thin Glass"
			
				# We will handle thin glass as a specular material with arch-glass, and single-face enabled.
			
				# Write out a glass medium
				medium_uid = get_and_incr_next_uid()
				
				medium_name = export_name + " medium"
				out.puts "	<medium>"
				out.puts "		<uid>#{medium_uid}</uid>"
				write_name_string(out, 2, "name", medium_name)
				out.puts "		<precedence>100</precedence>"
				out.puts "		<basic>"
				out.puts "			<ior>1.5</ior>"
				out.puts "			<cauchy_b_coeff>0.0</cauchy_b_coeff>"
				export_absorption(mat, out)
				out.puts "		</basic>"
				out.puts "	</medium>"
				out.puts ""
				
				# Write out a specular material with arch-glass enabled.
				
				mat_uid = get_and_incr_next_uid()
				
				out.puts "	<material>"
				out.puts "		<uid>#{mat_uid}</uid>"
				write_name_string(out, 2, "name", export_name)
				out.puts "		<specular>"
				out.puts "			<transparent>true</transparent>"
				out.puts "			<arch_glass>true</arch_glass>"
				out.puts "			<single_face>true</single_face>"
				out.puts "			<internal_medium_uid>#{medium_uid}</internal_medium_uid>"
				
				bump_map = export_texture(mat, mat.bump_map, out, generate_uvs)  if mat.bump_map?
				bump_shader = mat.bump_map.shader_code if mat.bump_shader?
				
				out.puts "		</specular>"
				out.puts "	</material>"
				out.puts ""
				
				@material_uids[[mat.mat, generate_uvs]] = mat_uid
				
			when "Exit Portal"
				mat_uid = get_and_incr_next_uid()
				out.puts "	<material>"
				out.puts "		<uid>#{mat_uid}</uid>"
				write_name_string(out, 2, "name", export_name)
				out.puts "		<exit_portal/>"
				out.puts "	</material>"
				out.puts ""
				
				@material_uids[[mat.mat, generate_uvs]] = mat_uid
				
			end  #end case type
		
		rescue IndigoInvalidTabulatedException
			UI.messagebox("Material #{mat.name} contains invalid tabulated spectrum data.  Export aborted.")
			raise
		end  #end begin
	end  #end material writer


	####
	def export_constant_wavelen_dep_param(rgb, param, out)

		case param.type
		when "SketchUp"
			export_rgb_spectrum(out, rgb)
		when "Constant"
			export_constant_wdp(param, out, true)
		else
			# Control flow can get here if the texture is invalid.  In this case just use the SketchUp colour.
			export_rgb_spectrum(out,rgb)
		end
	end

	
	#####
	def export_constant_wip(rgb,param,out)

		constant = 0.0
		case param.type
		when "SketchUp"
			# Export normalized rgb value multiplied by b parameter.
			avg_rgb = ((rgb[0]+rgb[1]+rgb[2])/3.0)/255.0
			constant = avg_rgb * param.b.to_f
			constant = "%.4f" %(constant)  #format to 4 decimal places
		when "Constant"
			constant = param.b.to_f
		end
		out.puts "				<constant>#{constant}</constant>"
	end

	###
	def export_absorption(mat,out)

		out.puts "				<absorption_coefficient>"
		abs = mat.absorption_map
		case abs.type
		when "SketchUp"
			sketchup_mat_rgb = mat.rgb
			opacity = mat.alpha
			alpha_factor = opacity*opacity*1000
			out.puts "				<constant>"
			out.puts "					<rgb>"
			out.puts "						<rgb>" + "%.4f" %((1.0-sketchup_mat_rgb[0]/255.0)*alpha_factor) + " "+ "%.4f" %((1.0-sketchup_mat_rgb[1]/255.0)*alpha_factor) + " "+ "%.4f" %((1.0-sketchup_mat_rgb[2]/255.0)*alpha_factor) + "</rgb>"
			out.puts "						<gamma>1.0</gamma>"
			out.puts "					</rgb>"
			out.puts "				</constant>"
		when "Constant"
			export_constant_wdp(abs, out, true)
		when "Shader"
			export_shader(abs.shader_code, out, "false")
		end
		out.puts "				</absorption_coefficient>"
	end

	#this function just uses a different tag then the previous one
	def export_coating_absorption(mat,out)

		out.puts "				<absorption>"
		abs = mat.absorption_map
		case abs.type
		when "SketchUp"
			sketchup_mat_rgb = mat.rgb
			opacity = mat.alpha
			alpha_factor = opacity*opacity*1000
			out.puts "				<constant>"
			out.puts "					<rgb>"
			out.puts "						<rgb>" + "%.4f" %((1.0-sketchup_mat_rgb[0]/255.0)*alpha_factor) + " "+ "%.4f" %((1.0-sketchup_mat_rgb[1]/255.0)*alpha_factor) + " "+ "%.4f" %((1.0-sketchup_mat_rgb[2]/255.0)*alpha_factor) + "</rgb>"
			out.puts "						<gamma>1.0</gamma>"
			out.puts "					</rgb>"
			out.puts "				</constant>"
		when "Constant"
			export_constant_wdp(abs,out,true)
		when "Shader"
			export_shader(abs.shader_code,out,"false")
		end
		out.puts "				</absorption>"
	end
	
	####
	def export_sss(mat,out)

		out.puts "				<scattering_coefficient>"
		param = mat.sss_map
		case param.type
		when "Constant"
			export_constant_wdp(param,out,true)
		when "Shader"
			export_shader(param.shader_code,out,"false")
		end
		out.puts "				</scattering_coefficient>"
	end

	####
	def export_constant_wdp(param, out, new_syntax)

		out.puts "					<constant>" if new_syntax #some parameters still use the old syntax (without the constant tag)
		case param.spectrum_type
		when "RGB"
			#p [param.rgb_r,param.rgb_g,param.rgb_b]
			gain=param.rgb_gain.to_f
			out.puts "					<rgb>"
			out.print "						<rgb>" + "%.4f" %(param.rgb_r.to_f*gain) + " "+ "%.4f" %(param.rgb_g.to_f*gain) + " "+ "%.4f" %(param.rgb_b.to_f*gain) + "</rgb>\n"
			out.print "						<gamma>#{param.rgb_gamma}</gamma>\n"
			out.print "					</rgb>\n"
		when "Blackbody"
			out.puts "					<blackbody>"
			out.puts "						<temperature>#{param.bb_temperature}</temperature>"
			out.puts "						<gain>#{param.bb_gain}</gain>"
			out.puts "					</blackbody>"
		when "Uniform"
			out.puts "					<uniform>"
			out.puts "						<value>#{param.uni_value}</value>"
			out.puts "					</uniform>"
		when "Tabulated"
			values=param.reg_values
			num_values=values.split(" ").length
			if num_values<2
				raise IndigoInvalidTabulatedException
			end
			start_wl=param.reg_start.to_f/1000.0
			end_wl=param.reg_end.to_f/1000.0
			out.puts "					<regular_tabulated>"
			out.puts "						<start_wavelength>#{start_wl}E-06</start_wavelength>"
			out.puts "						<end_wavelength>#{end_wl}E-06</end_wavelength>"
			out.puts "						<num_values>#{num_values}</num_values>"
			out.puts"						<values>#{values}</values>"
			out.puts "					</regular_tabulated>"
		end
		out.puts "					</constant>" if new_syntax
	end
	
	
	#############################
	# Returns mat_uid
	def export_front_or_back_diffuse_mat(mat, out, generate_uvs, is_front)
		mat_uid = get_and_incr_next_uid()
		
		out.puts "	<material>"
		out.puts "		<uid>#{mat_uid}</uid>"
		write_name_string(out, 3, "name", mat.name + " " + (is_front ? "front" : "back"))
		out.puts "		<diffuse>" 
		if mat.emitter?
			out.puts "			<layer>#{@used_layers_array.index(mat.layer.to_i)}</layer>" if !(@igm)
		end
		
		# Export all of the texture xml and collect the indices
		@tex_index = 0 # We are exporting a new material so reset the texture index.
		albedo_tex 		= export_texture(mat, mat.albedo_texture, out, generate_uvs) if mat.albedo_texture?		
		bump_map 		= export_texture(mat, mat.bump_map, out, generate_uvs)  if mat.bump_map?
		normal_map 		= export_texture(mat, mat.normal_map, out, generate_uvs)  if mat.normal_map?
		disp_map 		= export_texture(mat, mat.disp_map, out, generate_uvs)  if mat.disp_map?
		emission_map 	= export_texture(mat, mat.emission_map, out, generate_uvs) if mat.emission_map?

		# Collect all of the shaders
		albedo_shader 	= mat.albedo_texture.shader_code if mat.albedo_shader?
		bump_shader 	= mat.bump_map.shader_code if mat.bump_shader?
		normal_shader 	= mat.normal_map.shader_code if mat.normal_shader?
		disp_shader 	= mat.disp_map.shader_code if mat.disp_shader?
		emission_shader = mat.emission_map.shader_code if mat.emission_shader?
			
		sketchup_mat_rgb = mat.rgb
			
		export_wavelen_dep_param("albedo", sketchup_mat_rgb, albedo_tex, albedo_shader, mat.albedo_texture, out)
		
		# Export the common parameters
		export_wavelen_indep_param("bump", bump_map, bump_shader, mat.bump_map, out) if mat.use_bump?
		export_wavelen_indep_param("normal_map", normal_map, normal_shader, mat.normal_map, out) if mat.use_normal? 
		export_wavelen_indep_param("displacement", disp_map, disp_shader, mat.disp_map, out) if mat.use_displace?
		export_base_emission(sketchup_mat_rgb, mat.base_emission, out) if mat.emitter?
		export_wavelen_dep_param("emission", sketchup_mat_rgb, emission_map, emission_shader, mat.emission_map, out) if mat.use_emission?
					
		out.puts "		</diffuse>"
		out.puts "	</material>"
		out.puts ""
		return mat_uid
	end
	
	
	def export_simple_thin_mat(mat, out, export_name, generate_uvs)
					
		# Export front diffuse material
		front_mat_uid = export_front_or_back_diffuse_mat(mat, out, generate_uvs, true)
		
		# Export back diffuse material
		#back_mat_uid = export_front_or_back_diffuse_mat(mat, out, generate_uvs, false)
		
		mat_uid = get_and_incr_next_uid()
		
		# Export the double-sided thin material
		out.puts "	<material>"
		out.puts "		<uid>#{mat_uid}</uid>"
		write_name_string(out, 2, "name", export_name)
		out.puts "		<double_sided_thin>" 
		
		out.puts "			<front_material_uid>#{front_mat_uid}</front_material_uid>"
		out.puts "			<back_material_uid>#{front_mat_uid}</back_material_uid>"
		
		
		# Export all of the texture xml and collect the indices
		@tex_index = 0 # We are exporting a new material so reset the texture index.
		bump_map 				= export_texture(mat, mat.bump_map, out, generate_uvs)  if mat.bump_map?
		normal_map 				= export_texture(mat, mat.normal_map, out, generate_uvs)  if mat.normal_map?
		disp_map 				= export_texture(mat, mat.disp_map, out, generate_uvs)  if mat.disp_map? #and (!mat.clip_map?) #don't export the displacement map if this material has a clip map,  export it with the blend material
		transmittance_map 		= export_texture(mat, mat.albedo_texture, out, generate_uvs) if mat.albedo_texture? # export_texture(mat, mat.transmittance_map, out, generate_uvs) if mat.transmittance_map?
		r_f_map 				= export_texture(mat, mat.r_f_map, out, generate_uvs) if mat.r_f_map?
		front_fresnel_scale_map = export_texture(mat, mat.front_fresnel_scale_map, out, generate_uvs) if mat.front_fresnel_scale_map?
		back_fresnel_scale_map 	= export_texture(mat, mat.back_fresnel_scale_map, out, generate_uvs) if mat.back_fresnel_scale_map?
		front_roughness_map 	= export_texture(mat, mat.front_roughness_map, out, generate_uvs) if mat.front_roughness_map?
		back_roughness_map 		= export_texture(mat, mat.back_roughness_map, out, generate_uvs) if mat.back_roughness_map?
		
		# Collect all of the shaders
		bump_shader 			= mat.bump_map.shader_code if mat.bump_shader?
		normal_shader 			= mat.normal_map.shader_code if mat.normal_shader?
		disp_shader 			= mat.disp_map.shader_code if mat.disp_shader?
		transmittance_shader 	= mat.transmittance_map.shader_code if mat.transmittance_shader?
		r_f_shader 				= mat.r_f_map.shader_code if mat.r_f_shader?
		front_fresnel_scale_shader = mat.front_fresnel_scale_map.shader_code if mat.front_fresnel_scale_shader?
		back_fresnel_scale_shader = mat.back_fresnel_scale_map.shader_code if mat.back_fresnel_scale_shader?
		front_roughness_shader 	= mat.front_roughness_map.shader_code if mat.front_roughness_shader?
		back_roughness_shader 	= mat.back_roughness_map.shader_code if mat.back_roughness_shader?			
				
		sketchup_mat_rgb = mat.rgb
				
		export_wavelen_indep_param("r_f", r_f_map, r_f_shader, mat.r_f_map, out)
		export_wavelen_dep_param("transmittance", sketchup_mat_rgb, transmittance_map, transmittance_shader, mat.transmittance_map, out)
		export_wavelen_indep_param("front_roughness", front_roughness_map, front_roughness_shader, mat.front_roughness_map, out)
		export_wavelen_indep_param("back_roughness", back_roughness_map, back_roughness_shader, mat.back_roughness_map, out)
		export_wavelen_indep_param("front_fresnel_scale", front_fresnel_scale_map, front_fresnel_scale_shader, mat.front_fresnel_scale_map, out)
		export_wavelen_indep_param("back_fresnel_scale", back_fresnel_scale_map, back_fresnel_scale_shader, mat.back_fresnel_scale_map, out)
					
		# Export the common parameters
		export_wavelen_indep_param("bump", bump_map, bump_shader, mat.bump_map, out) if mat.use_bump?
		export_wavelen_indep_param("normal_map", normal_map, normal_shader, mat.normal_map, out) if mat.use_normal? 
		export_wavelen_indep_param("displacement", disp_map, disp_shader, mat.disp_map, out) if mat.use_displace?	

		
		out.puts "		</double_sided_thin>"
		out.puts "	</material>"
		out.puts ""
		
		@material_uids[[mat.mat, generate_uvs]] = mat_uid
		
		export_clip_map(mat, out, export_name, generate_uvs)  if mat.use_clip? #if mat.clip_map? or mat.clip_shader? #exports the clip map
	end

	#############################
	def export_blended_diffuse_transmitter(mat,export_name,out,generate_uvs)
	
		type = mat.type   
		sketchup_mat_rgb = mat.rgb 	
		mat_a = export_name + " diffuse transmitter"
		mat_b = export_name + " opaque"
		
		
		### Export material b (diffuse/phong material) ###
		mat_b_uid = get_and_incr_next_uid()
		
		out.puts "	<material>"
		out.puts "		<uid>#{mat_b_uid}</uid>"
		write_name_string(out, 2, "name", mat_b)
		out.print "			<#{mat.indigo_type}>\n" 
		if mat.emitter?
			out.puts "			<layer>#{@used_layers_array.index(mat.layer.to_i)}</layer>" if !(@igm)
			#@used_layers.insert(mat.layer)
		end

		# Export all of the texture xml and collect the indices
		albedo_tex 		= export_texture(mat, mat.albedo_texture, out, generate_uvs) if mat.albedo_texture?		
		bump_map 		= export_texture(mat, mat.bump_map, out, generate_uvs)  if mat.bump_map?
		normal_map 		= export_texture(mat, mat.normal_map, out, generate_uvs)  if mat.normal_map?
		exponent_map 	= export_texture(mat, mat.exponent_map, out, generate_uvs) if mat.exponent_map?
		disp_map 		= export_texture(mat, mat.disp_map, out, generate_uvs)  if mat.disp_map?
		emission_map 	= export_texture(mat, mat.emission_map, out, generate_uvs) if mat.emission_map?

		# Collect all of the shaders
		albedo_shader 	= mat.albedo_texture.shader_code if mat.albedo_shader?
		bump_shader 	= mat.bump_map.shader_code if mat.bump_shader?
		normal_shader 	= mat.normal_map.shader_code if mat.normal_shader?
		exponent_shader = mat.exponent_map.shader_code if mat.exponent_shader?
		disp_shader 	= mat.disp_map.shader_code if mat.disp_shader?
		emission_shader = mat.emission_map.shader_code if mat.emission_shader?
				
		elem_name = (type == "Phong") ? "diffuse_albedo" : "albedo"
		export_wavelen_dep_param(elem_name, sketchup_mat_rgb, albedo_tex, albedo_shader, mat.albedo_texture, out)

		if type == "Phong"
			out.puts "			<ior>#{mat.ior}</ior>"
			export_wavelen_indep_param("exponent", exponent_map, exponent_shader, mat.exponent_map, out)
		end

		# Export the common parameters
		export_wavelen_indep_param("bump", bump_map, bump_shader, mat.bump_map, out) if mat.use_bump?
		export_wavelen_indep_param("normal_map", normal_map, normal_shader, mat.normal_map, out) if mat.use_normal? 
		export_wavelen_indep_param("displacement", disp_map, disp_shader, mat.disp_map, out) if mat.use_displace?
		export_base_emission(sketchup_mat_rgb, mat.base_emission, out) if mat.emitter?
		export_wavelen_dep_param("emission", sketchup_mat_rgb, emission_map, emission_shader, mat.emission_map, out) if mat.use_emission?
					
		out.puts "		</#{mat.indigo_type}>"
		out.puts "	</material>"
		out.puts ""

		#Blend the material with a diffuse transmitter material
		
		
		#### Export material a, the diffuse transmitter material
		mat_a_uid = get_and_incr_next_uid()

		@tex_index = 0 # We are exporting a new material so reset the texture index			
		out.puts "	<material>"
		out.puts "		<uid>#{mat_a_uid}</uid>"
		write_name_string(out, 2, "name", mat_a)
		out.print "		<diffuse_transmitter>\n"
		if mat.emitter?
			out.puts "			<layer>#{@used_layers_array.index(mat.layer.to_i)}</layer>" if !(@igm)
			#@used_layers.insert(mat.layer)
		end

		# Export all of the texture xml and collect the indices
		albedo_tex 	= disp_map = emission_map = bump_map = nil
		albedo_shader = disp_shader = emission_shader = bump_shader = nil

		albedo_tex 		= export_texture(mat, mat.albedo_texture, out, generate_uvs) if mat.albedo_texture?
		bump_map 		= export_texture(mat, mat.bump_map, out, generate_uvs)  if mat.bump_map?
		normal_map 		= export_texture(mat, mat.normal_map, out, generate_uvs)  if mat.normal_map?	
		disp_map 		= export_texture(mat, mat.disp_map, out, generate_uvs)  if mat.disp_map?
		emission_map 	= export_texture(mat, mat.emission_map, out, generate_uvs) if mat.emission_map?

		# Collect all of the shaders
		albedo_shader 	= mat.albedo_texture.shader_code if mat.albedo_shader?
		bump_shader 	= mat.bump_map.shader_code if mat.bump_shader?
		normal_shader 	= mat.normal_map.shader_code if mat.normal_shader?
		disp_shader 	= mat.disp_map.shader_code if mat.disp_shader?
		emission_shader = mat.emission_map.shader_code if mat.emission_shader?

		
		export_wavelen_dep_param("albedo", sketchup_mat_rgb, albedo_tex, albedo_shader, mat.albedo_texture, out)

		# Export the common parameters
		export_wavelen_indep_param("bump", bump_map, bump_shader, mat.bump_map, out) if mat.use_bump?
		export_wavelen_indep_param("normal_map", normal_map, normal_shader, mat.normal_map, out) if mat.use_normal? 
		export_wavelen_indep_param("displacement", disp_map, disp_shader, mat.disp_map, out) if mat.use_displace?
		export_base_emission(sketchup_mat_rgb, mat.base_emission, out) if mat.emitter?
		export_wavelen_dep_param("emission", sketchup_mat_rgb, emission_map, emission_shader, mat.emission_map, out) if mat.use_emission?
		
		out.print "		</diffuse_transmitter>\n"
		out.print "	</material>\n"
		out.puts ""

		# Export the blended diffuse transmitter
		blend_mat_uid = get_and_incr_next_uid()
		
		out.puts "	<material>"
		out.puts "		<uid>#{blend_mat_uid}</uid>"
		write_name_string(out,2,"name",export_name)
		out.puts "		<blend>"
		out.puts "			<a_mat_uid>#{mat_a_uid}</a_mat_uid>"
		out.puts "			<b_mat_uid>#{mat_b_uid}</b_mat_uid>"
		out.puts "			<blend>"
		out.puts "				<constant>#{mat.alpha}</constant>"
		out.puts "			</blend>"
		out.puts "		</blend>"
		out.puts "	</material>"
		out.puts ""
		
		# Store in material hash.
		@material_uids[[mat.mat, generate_uvs]] = blend_mat_uid
	end

	####################################Write the Texture Information#################
	# tex is of type IndigoTexture
	# Returns texture index
	def export_texture(indigo_mat, tex, out, generate_uvs = false)

		su_mat = indigo_mat.mat  #get the SU material that we are exporting
		case tex.type
			when "SketchUp"
				tex_path = self.exported_texture_filename(su_mat)
			when "External"
				tex_path = tex.path
		end
		if not tex.valid?
			fixed_path = false
			if tex.type == "External"
				# Try and fix up the texture path by scanning the scene directory tree, looking for a matching file.
				result_path = SkIndigo.find_file_in_dir_tree(
					File.dirname(Sketchup.active_model.path), # Search directory
					File.basename(tex.path) # Target file name
				)
				#puts "Tried texture fixup, dir: '#{File.dirname(Sketchup.active_model.path)}', target: '#{File.basename(tex.path)}', result: '#{result_path}'"
				if result_path
					tex_path = result_path
					fixed_path = true
				end
			end
			
			if !fixed_path
				@export_warnings << "Could not find map '#{tex_path}' for material '#{su_mat.name}'"
				return
			end
		end	
		uvs = tex.uv_set_index()
		out.puts "				<texture>"
		out.print "					<uv_set_index>"
		if @igm or @preview_material
			out.print "#{tex.preview_uvset}"
		else
			out.print "#{uvs}"
		end
		out.puts "</uv_set_index>\n"
		
		if @igm
			out.puts "					<path>#{File.basename(tex_path)}</path>"   #the path for the texture is the file basename when exporting an IGM
		else
			out.puts "					<path>#{File.join(@tex_path, tex_path)}</path>" if tex.type=="SketchUp"
			out.puts "					<path>#{tex_path}</path>" if tex.type=="External"
		end
		
		out.puts "					<a>#{tex.a}</a>" if tex.a != 0.0 # Only write if a != default (default == 0.0)
		out.puts "					<b>#{tex.b}</b>" if tex.b != 1.0 # Only write if b != default (default == 1.0)
		out.puts "					<c>#{tex.c}</c>" if tex.c != 0.0 # Only write if c != default (default == 0.0)
		out.puts "					<exponent>#{tex.exponent}</exponent>"
		out.puts "					<smooth>#{tex.smooth?}</smooth>" if tex.smooth? # Only write if smooth != default (default == false)
		if generate_uvs
			texture = su_mat.texture
			if texture
				tex_height=texture.height.to_f
				tex_width=texture.width.to_f
			else
				tex_height=12.0  
				tex_width=12.0
			end
			out.puts "				<tex_coord_generation>"
			if generate_uvs=="Spherical"
				out.puts "				<spherical>"
				out.puts "					<rotation>"
				out.puts "						<axis_rotation>"
				out.puts "							<axis>0 0 1</axis>"
				out.puts "							<angle>#{@irs.exr_shift.to_f.degrees}</angle>"
				out.puts "						</axis_rotation>"
				out.puts "					</rotation>"
				out.puts "				</spherical>"
			else
				out.puts "					<uv>"
				out.puts "					<matrix>"
				out.puts "					#{1.0/tex_width} 0 0 #{1.0/tex_height}"
				out.puts "					</matrix>"
				out.puts "					<translation>"
				out.puts "					0 0"
				out.puts "					</translation>"
				out.puts "					</uv>"
			end
			out.puts "				</tex_coord_generation>"
		end
		out.puts "				</texture>"
		@textures.push([su_mat, tex])  #keep track of all the exported textures
		tex_index = @tex_index
		@tex_index += 1
		return tex_index  # Return the current index of this texture in the current material
	end

	#########################
	def export_texture_index(index,out)

		out.puts "				<texture>"
		out.puts "					<texture_index>#{index}</texture_index>"
		out.puts "				</texture>"
	end

	#########################
	def export_shader(shader_code,out,wld=nil)

		unless shader_code=~/CDATA/
			cdata=true
		else
			cdata=false
		end
		out.puts "				<shader>"
		if cdata
			out.puts "					<shader><![CDATA[#{shader_code}]]></shader>"
		else
			out.puts "					<shader>#{shader_code}</shader>"
		end
		out.puts "				<wavelength_dependent>#{wld}</wavelength_dependent>" if wld
		out.puts "				</shader>"
	end

	####################################Clip Map Export
	def export_clip_map(mat, out, clipped_name, generate_uvs)

		@tex_index = 0    # Reset the texture index variable since we are now exporting a new material
		clip_index = nil
		clip_shader = nil
		sketchup_mat_rgb = mat.rgb
		mat_name = mat.name
		mat_name = "previewmaterial" if mat == preview_material
		clip_map = mat.clip_map
		type = clip_map.type
		if type == "External"
			if not clip_map.valid?
				UI.messagebox ("Could not find clip map for material named #{mat.mat.name}") 
				return
			end	
		end
		
		write_null_mat(out) if !@null_exported
		
		
		mat_uid = get_and_incr_next_uid()
		
		out.print "		<material>\n"
		out.print "			<uid>#{mat_uid}</uid>\n"
		write_name_string(out, 3, "name", mat_name)
		out.print "			<blend>\n"
		write_name_string(out, 4, "a_name", clipped_name)
		out.print "				<b_name>None</b_name>\n"
		clip_index = export_texture(mat, clip_map, out, generate_uvs) if mat.clip_map?
		clip_shader = clip_map.shader_code if mat.clip_shader?
					
		out.puts "				<blend>"
		
		export_constant_wip(sketchup_mat_rgb, mat.clip_map,out) if !(clip_index or clip_shader)
		export_texture_index(clip_index, out) if clip_index
		export_shader(clip_shader, out) if clip_shader
		
		out.puts "				</blend>"
		out.puts "				<step_blend>true</step_blend>"  if (mat.clip_map? and clip_map.b.to_f == 1.0) # force step blend for clip maps
		out.print "			</blend>\n"
		out.print "		</material>\n"
		out.puts ""
		
		@material_uids[[mat.mat, generate_uvs]] = mat_uid
	end

	###############################################Write Textures########################
	def export_textures(textures)   #function takes an array of an array of SU materials and IndigoTexture objects
		
		return if textures.empty?
		
		Sketchup.active_model.start_operation("Export Textures")
		
		exported = Set.new # Set of exported-to pathnames
		tw = Sketchup.create_texture_writer
		entities = @model.active_entities
		helper_group = entities.add_group
		#p "number of textures found = #{textures.length}"
		for t in textures
			
			mat = t[0]  #get the SU material that references this texture
			tex = t[1]  #get the IndigoTexture object
			

			case tex.type
				when "SketchUp"  #export the texture to the textures directory
					if mat.respond_to?(:texture) and mat.texture !=nil
						export_filename = self.exported_texture_filename(mat)
						helper_group.material = mat
						
						if not exported.include?(export_filename)
							export_path = File.join(File.dirname(@path), tex_path, export_filename)
							
							if !FileTest.exist?(export_path) # Don't export if texture already exists on disk.
								tw.load(helper_group)   #load the texture if it has not yet been loaded
								status=tw.write(helper_group, export_path)
								exported.insert(export_filename)     #keep track of all of the exported texture filenames
							end
						end
					end		
				when "External"   #copy the file to the textures directory if we are exporting an IGM
					
					export_filename=File.basename(tex.path)
					export_path=File.join(File.dirname(@path),tex_path,export_filename)
					if @igm and (not exported.include?(export_filename))
						
						read_file=File.new((tex.path),"rb")
						contents=read_file.read
						read_file.close()
						write_file=File.new((export_path),"wb")
						write_file.write(contents)
						write_file.close()
						exported.insert(export_filename)   #keep track of all of the exported texture filenames
					end
				else
					puts("Texture not exported " + tex.inspect)
					 
			end
		end

		#entities.erase_entities helper_group
		Sketchup.active_model.abort_operation
		tw = nil
	end
	
	# See http://stackoverflow.com/questions/2174767/extracting-the-last-n-characters-from-a-ruby-string
	def get_last_n_chars(str, n)
		return str.reverse[0...n].reverse
	end
	
	def sanitise_string(str)
		return str.gsub(/[ \.\\\/:*#\[\]?"><\|]/, '_')
	end

	
	# Method to provide a unique exported texture filename for each material and keep the original filename extension (png, jpeg, tif)
	def exported_texture_filename(su_mat)
		if su_mat.texture != nil
		
			# Include the material colour in the filename, because the material colour affects the exported texture.
			col = su_mat.color
			param_digest = "#{col.red}_#{col.green}_#{col.blue}_#{col.alpha}"
			
			# Include some of the sketchup file path.  The idea here is to try and get a unique texture name, so that different sketchup files with e.g. textures called 'wood.jpg' don't interfere.
			model_path_prefix = get_last_n_chars(sanitise_string(@model.path.downcase), 30)

			texture_filename = File.basename(su_mat.texture.filename)
			
			filename_parts = texture_filename.split(".")
			
			if filename_parts.size < 2
				# Texture filename does not have the form 'name.extension'.  Possibly it is the empty string.
				# Just try and make a unique texture filename not using the texture filename.
				return model_path_prefix + "_" + sanitise_string(su_mat.name.downcase).slice(0..25) + "_" + param_digest + ".jpg"
			end
			
			extension = filename_parts.last.downcase
			
			if not %w/jpg png gif tiff tif/.include?(extension)
				extension = "jpg"
			end
			
			return [model_path_prefix + "_" + sanitise_string(get_last_n_chars(texture_filename, 20)) + "_" + sanitise_string(su_mat.name.downcase).slice(0..25) + "_" + param_digest, extension].join(".")
		else
			puts "Error! Could not get su_mat.texture"
			return nil
		end
	end


	###############Write the Null Material
	def write_null_mat(out)
		if not @null_exported
			out.puts "	<material>"
			out.puts "		<uid>#{get_and_incr_next_uid()}</uid>"
			out.puts "		<name>None</name>"
			out.puts "		<null_material/>"
			out.puts "	</material>"
			out.puts ""
			@null_exported = true
		end
	end

	############################################IGM Exporter
	def export_igm(mat,pack)

		out = SkIndigo.new_writable_utf8_file(path)
		#@tex_path=File.dirname(path)
		@tex_path=""
		@igm=true
		out.puts "<?xml version=\"1.0\" encoding=\"utf-8\"?>"
		out.puts "<scenedata>"
		
		begin
			export_material(mat,out)
		rescue
			out.close
			UI.messagebox "Errors occured while exporting the IGM file."
			@igm=false
			raise
		end
		out.puts "</scenedata>"
		out.close
		
		begin
			
			export_textures(@textures)  #export textures collected during the material export process
			
		rescue
			UI.messagebox "Errors occured while exporting the material's textures."
			raise
		end
		
		@igm=false	
		#######the IGM spec does not require exporting of nkdata files to the IGM path
		#begin
			#for data in @nkdata
				#nk=NKdata.new(data)
				#nk.export(File.dirname(@path))  #exports the nkdata to the textures directory
			#end
		#rescue
			#UI.messagebox "Errors occured while exporting the material's nkdata."
			#raise
		#end
		
		if pack   #the user wishes to pack the material
			status=SkIndigo.pack_igm(path)
			if status==nil
				UI.messagebox("Unable to pack IGM file.")
			end
		else
			UI.messagebox "Material saved to #{path}"
		end

	end

	#########################################Camera export
	def export_camera(v,out)

		aperture=@irs.aperture
		ap_shape=@irs.aperture_shape
		focal_point=@irs.focal_point
		exr_shift=@irs.exr_shift.to_f
		env=@irs.environment
		shutterSpeed=1/(@irs.shutter_speed.to_f)

		h = v.vpheight   #get the dimensions of the viewport
		w = v.vpwidth
		user_camera = v.camera
		f=user_camera.focal_length
		user_eye = user_camera.eye
		user_y = user_camera.yaxis #camera up vector
		user_z = user_camera.zaxis #camera forward vector
		sensor_width=user_camera.image_width.to_f/1000

		if SkIndigo.photomatch?(v)  #check if the current view is photomatched
			aspect=user_camera.aspect_ratio     #view will have a set aspect ratio if in photo match mode
			if sensor_width==0.0                
				sensor_distance=(f.to_f/aspect)/1000
				sensor_width=0.036
			else
				sensor_distance=f.to_f/1000
			end
		else
			sensor_width=0.036 if sensor_width==0.0
			old_aspect=user_camera.aspect_ratio
			if old_aspect==0.0  #if aspect ratio is not set, set it to the viewport aspect ratio
				old_aspect=(w.to_f)/(h.to_f)
				sensor_distance=(f.to_f/old_aspect)/1000
			else
				sensor_distance=f.to_f/1000
			end
			aspect=old_aspect
		end

		unless user_camera.perspective?
			camera_type="orthographic"
			pt=[0,0,0]
			viewport_aspect=(w.to_f)/(h.to_f)
			if viewport_aspect>aspect
				w=w*(aspect/viewport_aspect)
			end
			sensor_width=v.pixels_to_model(w,pt)*@scale
		else
			camera_type=@irs.camera_type
		end

		stop=aperture.split("\/")[1].to_f
		ap_radius=sensor_distance/(stop*2)

		#Convert the stored focal point to a focus distance
		if focal_point!="auto"
			vector=user_eye - focal_point if focal_point
			focus_distance=vector.length*@scale
		else
			focus_distance=2.0
		end

		out.puts "	<camera>"
		out.puts "		<uid>#{get_and_incr_next_uid()}</uid>"
		out.puts "		<camera_type>#{camera_type}</camera_type>"
		out.puts "		<pos>" + "%.4f" %(user_eye.x*@scale)+ " "+ "%.4f" % (user_eye.y*@scale) + " " + "%.4f" %(cz=user_eye.z*@scale)+ "</pos>"
		out.puts "		<up>" + "%.4f" %(user_y.x)+ " "+ "%.4f" % (user_y.y) + " " + "%.4f" %(user_y.z)+ "</up>"
		out.puts "		<forwards>" + "%.4f" %(user_z.x)+ " "+ "%.4f" % (user_z.y) + " " + "%.4f" %(user_z.z)+ "</forwards>"
		out.puts "		<aperture_radius>" + "%.4f" %(ap_radius)+"</aperture_radius>"
		out.puts "		<focus_distance>"+ "%.4f" %(focus_distance)+"</focus_distance>"
		out.puts "		<aspect_ratio>" + "%.4f" %(aspect) + "</aspect_ratio>"
		out.puts "		<sensor_width>#{sensor_width}</sensor_width>"
		out.puts "		<lens_sensor_dist>" + "%.4f" %(sensor_distance)+"</lens_sensor_dist>"
		out.puts "		<white_balance>#{@irs.white_balance}</white_balance>"
		out.puts "		<autofocus/>" if focal_point=="auto"
		out.puts "		<exposure_duration>#{shutterSpeed}</exposure_duration>"
		out.puts "		<aperture_shape>"
		case ap_shape
			when "circular"
				out.puts "			<circular/>"
			when "image"
				if !FileTest.exist?(@irs.aperture_image)
					UI.messagebox("Error: Aperture image path is invalid.")
					raise
				end
				out.puts "			<image>"
				out.puts "				<path>#{@irs.aperture_image}</path>"
				out.puts "			</image>"
			else
				out.puts "			<generated>"
				out.puts "				<num_blades>#{ap_shape}</num_blades>"
				out.puts "				<blade_offset>#{@irs.blade_offset}</blade_offset>"
				out.puts "				<blade_curvature_radius>#{@irs.blade_curvature_radius}</blade_curvature_radius>"
				out.puts "				<start_angle>#{@irs.start_angle.to_f*3.14159/180}</start_angle>"
				out.puts "			</generated>"
			
		end
		out.puts "		</aperture_shape>"
		if @irs.use_obstacle_map?
			if !FileTest.exist?(@irs.obstacle_map)
					UI.messagebox("Error: Aperture obstacle map path is invalid.")
					raise
			else
				out.puts "		<obstacle_map>"
				out.puts "			<path>#{@irs.obstacle_map}</path>"
				out.puts "		</obstacle_map>"
			end
		end


		out.puts "		<lens_shift_up_distance>#{@irs.shift_up.to_f/1000.0}</lens_shift_up_distance>"
		out.puts "		<lens_shift_right_distance>#{@irs.shift_right.to_f/1000.0}</lens_shift_right_distance>" 							
		out.puts "	</camera>"
		out.puts ""
	end
	
	
	
	####
	def export_metadata(out)

		out.puts "	<metadata>"
		out.puts "		<uid>#{get_and_incr_next_uid()}</uid>"
		out.puts "		<exporter>SkIndigo version:#{SkIndigo::VERSION}</exporter>"
		out.puts "		<platform>#{Object::RUBY_PLATFORM}</platform>"
		author = ENV["USERNAME"]
		if author == nil
			author = ENV["USER"]
		end
		out.puts "	    <author>#{reinterpret_as_utf8(author.to_s)}</author>"
		out.puts "		<created_date>#{Time.now}</created_date>"
		out.puts "		<exporter_specific>"
		out.puts "			<sketchup_version>#{Sketchup.version}</sketchup_version>"
		out.puts "		</exporter_specific>"
		out.puts "	</metadata>"
		out.puts ""
	end
	
	
	####
	def export_section_planes(out)

		ents=Sketchup.active_model.active_entities
		planes=ents.find_all {|e| e.class==Sketchup::SectionPlane}

		planes.each {|section_plane|

			if section_plane.get_attribute('indigo07_settings','active_section',false)
				plane=section_plane.get_plane
				point=Geom::Point3d.new(0,0,0)
				pp=point.project_to_plane(plane)
				normal=Geom::Vector3d.new(plane[0],plane[1],plane[2]).normalize
				out.puts "		<section_plane>"
				out.puts "		<point>#{pp.x.to_f*@scale} #{pp.y.to_f*@scale} #{pp.z.to_f*@scale}</point>"
				out.puts "		<normal>#{normal.x} #{normal.y} #{normal.z}</normal>"
				out.puts "		<enabled>true</enabled>"
				out.puts "		<cull_geometry>#{@irs.cull_geometry?}</cull_geometry>"
				out.puts "		</section_plane>"
			end #if
		}
	end
	
	
	###################################Export Render Settings
	def export_render_settings(out)

		mat_id = false
		alpha = false
		gpu = @irs.use_gpu?

		case @irs.method
			when "MLT"
				met=true
				bidir=false
				hybrid=false
			when "PT"
				met=false
				bidir=false
				hybrid=false
			when "Hybrid"
				met=false
				bidir=false
				hybrid=true
			when "Bidir MLT"
				met=true
				bidir=true
				hybrid=false
			when "Bidir PT"
				met=false
				bidir=true
				hybrid=false
			when "Bidir Hybrid"
				met=false
				bidir=true
				hybrid=true
			when "Alpha"  #for backwards compatibility with previous skindigo versions - alpha used to be a tracing method
				@irs.method="Bidir PT"
				@irs.foreground_alpha=true
			when "Material_ID"
				met=false
				bidir=false
				hybrid=false
				mat_id=true
				gpu=false
		end

		width=@irs.width.to_i
		height=@irs.height.to_i
		view=Sketchup.active_model.active_view
		current_camera=view.camera
		old_aspect=current_camera.aspect_ratio

		if old_aspect==0.0 
			old_aspect=view.vpwidth.to_f/view.vpheight.to_f
			height=(width.to_f/old_aspect).to_i    #set the height to match the aspect ratio of the viewport
			#width=(height.to_f*old_aspect).to_i    #set the width to match the aspect ratio of the viewport
			
		end

		if SkIndigo.photomatch?(Sketchup.active_model.active_view)   
			current_aspect=current_camera.aspect_ratio   #this is the aspect ratio of the photomatch
			aspect=width.to_f/height.to_f
			if (current_aspect-aspect).abs>0.01  #check if rendered image aspect ratio matches photomatch aspct ratio
				height=(width/current_aspect).to_i   #set the height to match the aspect ratio of the photo
				UI.messagebox "Image dimensions set to #{width}x#{height} to match photo aspect ratio."
			end
		end

		out.puts "	<renderer_settings>"
		out.puts "		<uid>#{get_and_incr_next_uid()}</uid>"
		out.puts "		<width>#{width}</width>"
		out.puts "		<height>#{height}</height>"
		out.puts "		<bidirectional>true</bidirectional>" if bidir
		out.puts "		<bidirectional>false</bidirectional>" if not bidir
		out.puts "		<metropolis>true</metropolis>" if met
		out.puts "		<metropolis>false</metropolis>" if not met
		out.puts "		<hybrid>true</hybrid>" if hybrid
		out.puts "		<hybrid>false</hybrid>" if not hybrid
		out.puts "		<halt_time>#{@irs.halt}</halt_time>" if @irs.halt_unit=="sec"
		out.puts "		<halt_samples_per_pixel>#{@irs.halt}</halt_samples_per_pixel>" if @irs.halt_unit=="spp"
		out.puts "		<image_save_period>3.0</image_save_period>" if @preview_scene
		out.puts "		<image_save_period>#{@irs.image_save_period}</image_save_period>" if !@preview_scene
		out.puts "		<auto_choose_num_threads>true</auto_choose_num_threads>" if @irs.auto_choose_threads?
		out.puts "		<auto_choose_num_threads>false</auto_choose_num_threads>" if !@irs.auto_choose_threads?
		out.puts "		<num_threads>#{@irs.num_threads}</num_threads>" if !@irs.auto_choose_threads?
		out.puts "		<save_igi>true</save_igi>" if @irs.save_igi?
		out.puts "		<save_igi>false</save_igi>" if !@irs.save_igi?
		out.puts "		<watermark>true</watermark>" if @irs.watermark?
		out.puts "		<watermark>false</watermark>" if !@irs.watermark?
		out.puts "		<aperture_diffraction>#{@irs.aperture_diffraction?}</aperture_diffraction>"
		out.puts "		<render_foreground_alpha>#{@irs.foreground_alpha?}</render_foreground_alpha>"
		out.puts "		<material_id_tracer>true</material_id_tracer>" if mat_id
		out.puts "		<material_id_tracer>false</material_id_tracer>" if not mat_id
		out.puts "		<info_overlay>true</info_overlay>" if @irs.info_overlay?
		out.puts "		<info_overlay>false</info_overlay>" if !@irs.info_overlay?
		out.puts "		<glass_acceleration>#{@irs.glass_acceleration?}</glass_acceleration>"
		out.puts "		<super_sample_factor>#{@irs.supersample_factor}</super_sample_factor>"
		out.puts "		<vignetting>#{@irs.vignetting?}</vignetting>"
		out.puts "		<clamp_contributions>#{@irs.clamp_contributions?}</clamp_contributions>"
		out.puts "		<max_contribution>#{@irs.max_contribution}</max_contribution>"
		out.puts "		<gpu_max_depth>#{@irs.gpu_max_depth}</gpu_max_depth>"
		#out.puts "		<frame_upload_period>#{irs.frame_upload_period}</frame_upload_period>" #feature no longer used as of Indigo 3.0
		out.puts "		<splat_filter>"
		out.puts "			<fastbox/>"
		out.puts "		</splat_filter>"
		#out.puts "		<max_depth>2000</max_depth>"
		#out.puts "		<post_process_diffraction>false</post_process_diffraction>"    #this feature currently disabled
		out.puts "		<gpu>#{gpu}</gpu>"
		irs_model=IndigoRenderSettings.new
		if @irs.render_region?
			#ensure we get region dimensions from model attributes, not scene attributes
			region=irs_model.region
			if (region[1].to_i>0) and (region[3].to_i>0)
				out.puts "		<render_region>"
				out.puts "			<x1>#{[region[0].to_i,width].min}</x1>"
				out.puts "			<y1>#{[region[2].to_i,height].min}</y1>"
				out.puts "			<x2>#{[region[1].to_i,width].min}</x2>"
				out.puts "			<y2>#{[region[3].to_i,height].min}</y2>"
				out.puts "		</render_region>"
			end
		end
		#out.puts "				<verbose>false</verbose>"
		out.puts "	</renderer_settings>"
		out.puts ""
	end

	
	######
	def get_used_layers()
		Sketchup.active_model.materials.each {|mat|
			im=IndigoMaterial.new(mat)
			layer=im.layer
			if layer!=-1
				@used_layers_set.insert(im.layer)
			end
		}
		irs=IndigoRenderSettings.new
		case irs.environment
			when "Sun and Sky"
				@used_layers_set.insert(@irs.sun_layer.to_i)
				@used_layers_set.insert(@irs.sky_layer.to_i)
			when "Environment Map"
				env_map_layer = @irs.env_map_layer.to_i
				if env_map_layer == -1
					env_map_layer = 0
				end
				@used_layers_set.insert(env_map_layer)
				if (@irs.sunsky_weight.to_f>0.0) and (@irs.bg_sun?)
					@used_layers_set.insert(@irs.sun_layer.to_i)
				end
				#@used_layers.insert(@irs.sky_layer.to_i) if @irs.bg_sun?
		end
				
		@used_layers_array = @used_layers_set.to_a
	end


	####
	def export_light_layer_names(out)

		irs_model=IndigoRenderSettings.new
		#export only used layers
		layer_names=irs_model.light_layer_names
		@used_layers_array.each_index {|index|
			this_layer=@used_layers_array[index].to_i
			

			export_layer_name(out,layer_names[this_layer],index)
		}
	end
	
	
	#################################################Export Tonemapping
	def export_tonemapping(out)

		out.puts "	<tonemapping>"
		out.puts "		<uid>#{get_and_incr_next_uid()}</uid>"
		case @irs.tm
			when "Reinhard"
				out.puts "		<reinhard>"
				out.puts "			<pre_scale>#{@irs.pre_scale}</pre_scale>"
				out.puts "			<post_scale>#{@irs.post_scale}</post_scale>"
				out.puts "			<burn>#{@irs.burn}</burn>"
				out.puts "		</reinhard>"
			when "Linear"
				out.puts "		<linear>"
				out.puts "			<scale>#{@irs.linear_scale}</scale>"
				out.puts "		</linear>"
			when "Filmic"
				out.puts "		<filmic>"
				out.puts "			<scale>#{@irs.filmic_scale}</scale>"
				out.puts "		</filmic>"
			when "Camera"
				out.puts "		<camera>"
				out.puts "			<film_iso>#{@irs.iso}</film_iso>"
				out.puts "			<ev_adjust>#{@irs.ev_adjust}</ev_adjust>"
				out.puts "			<response_function_path>data/camera_response_functions/#{@irs.cam_resp_func}.txt</response_function_path>"
				out.puts "		</camera>"
					
		end
		out.puts "	</tonemapping>"
		out.puts ""
	end

	#################################Export Environment
	def export_environment(out)

		# Get the sun direction
		override = @irs.sun_pos_override?
		if override
			# Ensure we get the sun position from the model attributes, not the scene attributes
			irs_model = IndigoRenderSettings.new
			v = irs_model.sun_dir
			if v == []
				UI.messagebox("Warning: Sun position override not set.  Exporting SketchUp sun position.")
				override = false
			end
		end
		if !override
			si = Sketchup.active_model.shadow_info
			v = si["SunDirection"]
		end
		
		# Burn some UIDs, as as skylight etc.. will be converted to a SceneNodeBackgroundSettings node, and we don't want to clash with existing nodes.
		get_and_incr_next_uid()
		get_and_incr_next_uid()
		get_and_incr_next_uid()

		sun_layer = @irs.sun_layer
		sky_layer = @irs.sky_layer
				
		case @irs.environment
		when "Sun and Sky"
			
			@light_source = true
			#export_layer_name(out,"Sun",SUN_LAYER_INDEX)
			#export_layer_name(out,"Sky",SKY_LAYER_INDEX)
		
			out.puts "	<skylight>"
			out.puts "		<uid>#{get_and_incr_next_uid()}</uid>"
			out.puts "		<sundir>"+ "%.4f" % (v.x) + " " + "%.4f" % (v.y) + " " + "%.4f" % (v.z) + "</sundir>"
			out.puts "		<turbidity>#{@irs.turbidity}</turbidity>"
			out.puts "		<sun_layer>#{@used_layers_array.index(sun_layer.to_i)}</sun_layer>"
			out.puts "		<sky_layer>#{@used_layers_array.index(sky_layer.to_i)}</sky_layer>"
			out.puts "		<model>#{@irs.sky_model}</model>"
			out.puts "	</skylight>"
			out.puts ""
			#@used_layers.insert(sun_layer)
			#@used_layers.insert(sky_layer)
		when "Background"
			@light_source=true
			#export_layer_name(out,"Background",BACKGROUND_LAYER_INDEX)
			bg_gain = 20            #calibrated gain of background spectrum to balance meshlights of common efficacy scale
			sketchup_mat_rgb = Sketchup.active_model.rendering_options["BackgroundColor"]
			if sketchup_mat_rgb
				red = sketchup_mat_rgb.red*bg_gain
				green = sketchup_mat_rgb.green*bg_gain
				blue = sketchup_mat_rgb.blue*bg_gain
			else
				red = 0.0
				green = 0.0
				blue = 0.0
			end
			out.puts "	<background>"
			out.puts "		<uid>#{get_and_incr_next_uid()}</uid>"
			out.puts "		<spectrum>"
			out.puts "			<rgb>"
			out.puts "				<rgb>#{red} #{green} #{blue}</rgb>"
			out.puts "				<gamma>2.2</gamma>"
			out.puts "			</rgb>"
			out.puts "		</spectrum>"
			out.puts "	</background>"
			out.puts ""
		when "Environment Map"
			
			begin
				envmap_angle = @irs.exr_shift.to_f
				mats = Sketchup.active_model.materials
				# Locate the background material
				bg_mat = mats.find {|mat| mat.name == @irs.bg_material}
				if !bg_mat
					UI.messagebox "Warning: Background material not found.  You must specify a valid material name to use as an environment map."
					raise "Invalid_Background_Material"
				end
								
				bg_mat = IndigoMaterial.new(bg_mat)
				if bg_mat.layer == -1
					puts "Note: Environment material light layer was off, Setting to 0."
					bg_mat.layer = 0
				end				
				
				@light_source = true

				# Export the sun_sky material
				if @irs.bg_sun?
					@light_source=true
					bg_mat_name="background"
					export_material(bg_mat, out, "Spherical") #first export the background material
					out.puts "	<material>"
					out.puts "		<uid>#{get_and_incr_next_uid()}</uid>"
					out.puts "		<name>sunsky</name>"
					out.puts "		<sunsky>"
					out.puts "			<sundir>"+ "%.4f" % (v.x) + " " + "%.4f" % (v.y) + " " + "%.4f" % (v.z) + "</sundir>"
					out.puts "			<turbidity>#{@irs.turbidity}</turbidity>"
					out.puts "			<enable_sky>false</enable_sky>"
					out.puts "			<sun_layer>#{@used_layers_array.index(sun_layer.to_i)}</sun_layer>" if @irs.sunsky_weight.to_f > 0.0
					#out.puts "			<sky_layer>#{@used_layers.index(sky_layer.to_i)}</sky_layer>" if @irs.bg_sun?
					out.puts "		</sunsky>"
					out.puts "	</material>"
					out.puts ""
					out.puts "	<background_settings>"
					out.puts "		<uid>#{get_and_incr_next_uid()}</uid>"
					out.puts "		<background_material>"
					out.puts "			<material>"
					out.puts "				<uid>#{get_and_incr_next_uid()}</uid>"
					write_name_string(out, 4, "name", bg_mat_name)
					out.puts "				<sum>"
					out.puts "					<mat>"
					out.puts "						<mat_name>sunsky</mat_name>"
					out.puts "						<weight>"
					out.puts "							<constant>#{@irs.sunsky_weight}</constant>"
					out.puts "						</weight>"
					out.puts "					</mat>"
					out.puts "					<mat>"
					write_name_string(out, 6, "mat_name", bg_mat.name)
					out.puts "						<weight>"
					out.puts "							<constant>#{@irs.bg_weight}</constant>"
					out.puts "						</weight>"
					out.puts "					</mat>"
					out.puts "				</sum>"
					out.puts "			</material>"
					out.puts "		</background_material>"
					out.puts ""
				else
					
					bg_mat_name = bg_mat.name
					out.puts "	<background_settings>"
					out.puts "		<uid>#{get_and_incr_next_uid()}</uid>"
					out.puts "		<background_material>"
					export_material(bg_mat, out, "Spherical") #export the background material
					#out.puts "	<name>#{bg_mat_name}</name>"
					out.puts "		</background_material>"
				end
					
				out.puts "		<emission_scale>"
				write_name_string(out, 3, "material_name", bg_mat_name)
				out.puts "			<measure>#{bg_mat.emission_measure}</measure>"
				out.puts "				<value>#{bg_mat.emission_scale}</value>"
				out.puts "		</emission_scale>"
				out.puts "	</background_settings>"
				out.puts ""
			rescue
				puts "Exception was thrown: #{$!}"
				# An exception was thrown, due to e.g. a missing background material.  Just use sun/sky.
				out.puts "<skylight>"
				out.puts "	<sundir>0 0.8 1</sundir>"
				out.puts "	<turbidity>2</turbidity>"
				out.puts "	<model>captured-simulation</model>"
				out.puts "</skylight>"
			end
			#@exported_mats=[]  #clear the array of exported materials so that the bg material will be exported again as a regular material
		end #end case
		
		# Burn some UIDs, as as skylight etc.. will be converted to a SceneNodeBackgroundSettings node, and we don't want to clash with existing nodes.
		get_and_incr_next_uid()
		get_and_incr_next_uid()
		get_and_incr_next_uid()
	end

	#####
	def export_layer_name(out,name,index)

		out.puts "	<layer_name>"
		write_name_string(out, 2, "layer_name", name)
		out.puts "		<layer_index>#{index}</layer_index>"
		out.puts "	</layer_name>"
		out.puts ""
	end

	##############################################Export RGB Spectrum
	def export_rgb_spectrum(out,sketchup_mat_rgb)

		out.puts "				<constant>"
		out.puts "					<rgb>"
		out.puts "						<rgb>" + "%.4f" %((sketchup_mat_rgb[0]/255.0)) + " "+ "%.4f" % ((sketchup_mat_rgb[1]/255.0)) + " "+ "%.4f" % ((sketchup_mat_rgb[2]/255.0)) + "</rgb>"
		out.puts "						<gamma>2.2</gamma>"
		out.puts "					</rgb>"
		out.puts "				</constant>"
	end

	###############################################Export Default Material
	def export_default_mat(out)
		#puts "export_default_mat()"
		
		@default_mat_uid = get_and_incr_next_uid

		colour = Sketchup.active_model.rendering_options["FaceFrontColor"]
		sketchup_mat_rgb = colour.red * 0.8, colour.green * 0.8, colour.blue * 0.8
		sketchup_mat_rgb = [216,216,216] if @monochrome=="Clay"
		out.puts "	<material>"
		out.puts "		<uid>#{@default_mat_uid}</uid>"
		out.puts "		<name>Default</name>"
		out.puts "		<oren_nayar>"
		out.puts  "			<albedo>"
		export_rgb_spectrum(out, sketchup_mat_rgb)
		out.puts "			</albedo>"
		out.puts "			<sigma>"
		out.puts "				<constant>0.5</constant>"
		out.puts "			</sigma>"
		out.puts "		</oren_nayar>"
		out.puts " 	</material>"
		out.puts ""
	end
	

	##############################################Fog Export
	def export_fog(out)

		options = model.rendering_options
		model_extents = model.bounds
		eye = model.active_view.camera.eye
		model_extents.add eye # Ensures that the camera is inside the fog sphere
		center = model_extents.center

		cx = center.x * @scale
		cy = center.y * @scale
		cz = center.z * @scale

		depth = model_extents.depth * @scale
		height = model_extents.height * @scale
		width = model_extents.width * @scale

		d = depth
		d = height if height > depth
		d = width if (width > height) and (width > depth)

		r = d / 2 #fog sphere radius

		fog_end = options["FogEndDist"]
		fog_end = fog_end * @scale
		fog_color = options["FogColor"]
		alpha_factor = 1

		sss = 2.772 / fog_end  # Distance at which light intensity is halved is 0.693/scattering coefficient

		out.puts "	<medium>"
		out.puts "		<uid>#{get_and_incr_next_uid()}</uid>"
		out.print "		<name>Fog_medium</name>\n"
		out.print "		<precedence>2</precedence>\n"
		out.print "		<basic>\n"
		out.print "			<ior>1.0</ior>\n"
		out.print "			<cauchy_b_coeff>0.0</cauchy_b_coeff>\n"
		out.puts "			<absorption_coefficient_spectrum>"
		out.puts "				<rgb>"
		out.print "					<rgb>#{(1.0-fog_color.red/255.0)*alpha_factor} #{(1.0-fog_color.green/255.0)*alpha_factor} #{(1.0-fog_color.blue/255.0)*alpha_factor}</rgb>\n"
		out.print "					<gamma>1.0</gamma>\n"
		out.print "				</rgb>\n"
		out.print "			</absorption_coefficient_spectrum>\n"
		out.print "			<subsurface_scattering>\n"
		out.print "				<scattering_coefficient_spectrum>\n"
		out.print "					<uniform>\n"
		out.print "						<value>#{sss}</value>\n"
		out.print "					</uniform>\n"
		out.print "				</scattering_coefficient_spectrum>\n"
		out.print "				<phase_function>\n"
		out.print "					<uniform/>\n"
		out.print "				</phase_function>\n"
		out.print "			</subsurface_scattering>\n"
		out.print "		</basic>\n"
		out.print "	</medium>\n"
		out.puts ""
		out.print "	<material>\n"
		out.puts "		<uid>#{get_and_incr_next_uid()}</uid>"
		out.print "		<name>Fog</name>\n"
		out.print "		<specular>\n"
		out.print "			<transparent>true</transparent>\n"
		out.print "		<internal_medium_name>Fog_medium</internal_medium_name>\n"
		out.print "		</specular>\n"
		out.print "	</material>\n"
		out.puts ""
		out.puts "	<sphere>"
		out.puts "		<uid>#{get_and_incr_next_uid()}</uid>"
		out.puts "		<center>#{cx} #{cy} #{cz}</center>"
		out.puts "		<radius>#{r}</radius>"
		out.puts "		<material_name>Fog</material_name>"
		out.puts "	</sphere>"
		out.puts ""
	end

	###########################################Export Sphere Primitivies
	def export_sphere(inst, out)

		trans = inst.transformation
		sphere_mat = inst.material

		if sphere_mat
			sphere_mat = IndigoMaterial.new(sphere_mat)
			sphere_mat_name = sphere_mat.name
		else
			sphere_mat_name = "Default"
		end

		box = inst.su_instance.bounds
		center = box.center.transform(trans) # Get the sphere in world coordinates
		cx = center.x * @scale
		cy = center.y * @scale
		cz = center.z * @scale

		r = box.height * @scale / 2 # Set radius of sphere equal to the half the height of its bounding box

		out.puts "	<sphere>"
		out.puts "		<uid>#{get_and_incr_next_uid()}</uid>"
		out.puts "		<center>#{cx} #{cy} #{cz}</center>"
		out.puts "		<radius>#{r}</radius>"
		write_name_string(out, 2, "material_name", sphere_mat_name)
		out.puts "	</sphere>"
		out.puts ""
	end


	$image_export_counter = 0


	#####returns true if the path is an absolute windows path
	def is_windows_path(path)
		# p "path[1,1] == #{path[1,1]}"
		# path[1] returns fixnum in ruby 1.8, so we need [1,1] for compatibility with never versions.
		if path[1,1] == ':'
			return true
		else
			return false
		end
	end


	#####returns a usable basename for windows paths on unix systems
	def basename_for_win_path(path)
		# Remove drive letter and colon.
		path_without_drive_letter = path[2..-1]
		# Replace backslash with forwards slash.
		path_with_coverted_slahes = path_without_drive_letter.gsub /\\+/, '/' #' for notepad++
		# Get the basename for the converted path.
		return File.basename(path_with_coverted_slahes)
	end


	#####get image basename, handling windows paths properly on unix.
	def image_basename(path)
		if SkIndigo.on_mac? and is_windows_path(path)
			# p "path is windows path"
			return basename_for_win_path(path)
		else
			# p "path is unix path"
			return File.basename(path)
		end
	end
	
	
	# Returns [material UID, material name] pair.
	def export_material_for_image(su_image, out)
	
		id = su_image.entityID.to_s
		path = su_image.path
		
		# The image component original path could be a windows path.
		# In that case, File.basename() will not give the expected result on unixy systems,
		# but instead just return the whole windows path.
		# image_basename() handles this a little better.
		basename = image_basename(path)
		# p "Image path: #{path}"
		# p "Image basename: #{basename}"

		found_texture=true
		unless FileTest.exist?(path)

			# Write the image to disk
			path = File.join(File.dirname(@path), tex_path, basename)

			# p "Writing image to '#{path}'..."

			tw = Sketchup.create_texture_writer
			tw.load(su_image)	
			status = tw.write(su_image, path)

			# TODO: handle status != 0?
		end
		mat_name = basename + id + "_" + $image_export_counter.to_s
		$image_export_counter += 1

		mat_uid = get_and_incr_next_uid()
		
		out.puts "	<material>"
		out.puts "		<uid>#{mat_uid}</uid>"
		write_name_string(out, 3, "name", mat_name)
		out.puts "		<diffuse>"
		if found_texture==true
			out.puts "			<texture>"
			out.puts "					<uv_set>default</uv_set>"
			out.puts "					<path>#{path}</path>"
			out.puts "					<a>0.0</a>"
			out.puts "					<b>1.0</b>"
			out.puts "					<c>0.0</c>"
			out.puts "					<exponent>2.2</exponent>"
			out.puts "				</texture>"
		end

		out.puts "			<albedo>"
		if found_texture==true
			out.puts "				<texture>"
			out.puts "					<texture_index>0</texture_index>"
			out.puts "				</texture>"
		else
			out.puts "					<constant>"
			out.puts "					<uniform>"
			out.puts "						<value>0.5</value>"
			out.puts "					</uniform>"
			out.puts "					</constant>"
		end
		out.puts "			</albedo>"
		out.puts "		</diffuse>"
		out.puts "	</material>"
		out.puts ""
	
		return [mat_uid, mat_name]
	end


	#####export sketchup images
	# image is a Sketchup::Image, Sketchup::Transformation pair.
	def export_image(image, out)
	
		su_image = image[0]
		trans = image[1]
		
		id = su_image.entityID.to_s
		path = su_image.path
		
		normal = su_image.normal
		origin = su_image.origin
		zrot = su_image.zrotation
		
		if SkIndigo.sketchup_is_2014_or_later
			trans = trans * su_image.transformation
			
			# Compute width and height.  We use pixelwidth here, as the overall transformation of the image will convert from pixel space to world space.
			width = su_image.pixelwidth.to_f * @scale # Compute width in metres
			height = su_image.pixelheight.to_f * @scale # Compute height in metres.
		else
			local_trans = Geom::Transformation.new(origin, normal)
			rot_trans = Geom::Transformation.rotation([0,0,0], Z_AXIS, zrot)
			trans = trans * (local_trans * rot_trans)
			
			width = su_image.width.to_f * @scale
			height = su_image.height.to_f * @scale
		end
		
		
		mat_uid, mat_name = export_material_for_image(su_image, out)
		
		# Write a mesh for the image
		mesh_uid = get_and_incr_next_uid()
		
		out.puts "	<mesh>"
		out.puts "		<uid>#{mesh_uid}</uid>"
		write_name_string(out, 2, "name", mat_name)
		out.puts " 		<embedded_2>"
		write_name_string(out, 3, "used_material_name", mat_name)
		#out.puts "				<v p=\"#{origin.x.to_f} #{origin.y.to_f} #{origin.z.to_f}\" n=\"#{normal.x.to_f} #{normal.y.to_f} #{normal.z.to_f}\"\/>"
		#out.puts "				<v p=\"#{origin.x.to_f+width} #{origin.y.to_f} #{origin.z.to_f}\" n=\"#{normal.x.to_f} #{normal.y.to_f} #{normal.z.to_f}\"\/>"
		#out.puts "				<v p=\"#{origin.x.to_f} #{origin.y.to_f+height} #{origin.z.to_f}\" n=\"#{normal.x.to_f} #{normal.y.to_f} #{normal.z.to_f}\"\/>"
		#out.puts "				<v p=\"#{origin.x.to_f+width} #{origin.y.to_f+height} #{origin.z.to_f}\" n=\"#{normal.x.to_f} #{normal.y.to_f} #{normal.z.to_f}\"\/>"
		out.puts "			<v p=\"0.0 0.0 0.0\" n=\"#{normal.x.to_f} #{normal.y.to_f} #{normal.z.to_f}\"\/>"
		out.puts "			<v p=\"#{width} 0.0 0.0\" n=\"#{normal.x.to_f} #{normal.y.to_f} #{normal.z.to_f}\"\/>"
		out.puts "			<v p=\"0.0 #{height} 0.0\" n=\"#{normal.x.to_f} #{normal.y.to_f} #{normal.z.to_f}\"\/>"
		out.puts "			<v p=\"#{width} #{height} 0.0\" n=\"#{normal.x.to_f} #{normal.y.to_f} #{normal.z.to_f}\"\/>"

		out.puts "			<uvs uv0=\"0.0000 0.0000\" \/>"
		out.puts "			<uvs uv0=\"1.0000 0.0000\" \/>"
		out.puts "			<uvs uv0=\"0.0000 1.0000\" \/>"
		out.puts "			<uvs uv0=\"1.0000 1.0000\" \/>"
		out.puts "			<t v=\"0 1 2\" uv=\"0 1 2\" m=\"0\"\/>"
		out.puts "			<t v=\"1 3 2\" uv=\"1 3 2\" m=\"0\"\/>"
		out.puts "		</embedded_2>"
		out.puts "	</mesh>"
		out.puts ""
		
		# Write a model for the image
		model_uid = get_and_incr_next_uid()
		out.puts "	<model>"
		out.puts "		<uid>#{model_uid}</uid>"
		self.write_transformation(trans, out)
		write_name_string(out, 2, "mesh_name", mat_name)
		out.puts "	</model>"
		out.puts ""
	end
	
	
	# Returns mesh UID
	def export_mesh(imesh, meshbuilder, mesh_dir, out)
	
		mesh_uid = get_and_incr_next_uid

		imesh.uid = mesh_uid
		
		#puts "Exporting mesh, assigned UID #{mesh_uid}"

		#Sketchup.set_status_text("Exporting Meshes...")
		
		out.puts "	<mesh>"
		out.puts "		<uid>#{mesh_uid}</uid>"
		write_name_string(out, 2, "name", imesh.name)
		
		if imesh.max_subdiv > 0 # Only export subdiv settings if they will have any effect (e.g. if subdiv is on)
			out.puts "		<max_num_subdivisions>#{imesh.max_subdiv}</max_num_subdivisions>" # This has default value zero so only needs to be exported if > 0.
			out.puts "		<subdivide_pixel_threshold>#{imesh.pixel_threshold}</subdivide_pixel_threshold>"
			out.puts "		<subdivide_curvature_threshold>#{imesh.curvature_threshold}</subdivide_curvature_threshold>"
			out.puts "		<subdivision_smoothing>#{imesh.subdivision_smoothing?}</subdivision_smoothing>"
			out.puts "		<displacement_error_threshold>0.0005</displacement_error_threshold>"
		end
		# out.puts "		<merge_vertices_with_same_pos_and_normal>true</merge_vertices_with_same_pos_and_normal>"   # merge_vertices_with_same_pos_and_normal has default value true
		out.puts "		<external>"
		if imesh.use_external?
			mesh_path = imesh.external_mesh_path
		else
			mesh_path = File.join(mesh_dir, imesh.guid + (imesh.flattened ? "_1_flattened" : "") + ".igmesh")
			
			# If this is the top-level geometry mesh, can't do caching with it yet.
			if imesh.top_level_geom_mesh? || !FileTest.exists?(mesh_path)
				imesh.write_to_file(mesh_path)
			end
		end
		out.puts "			<path>#{mesh_path}</path>"
		out.puts "		</external>"
		out.puts "	</mesh>"
		out.puts ""
		
		return mesh_uid
	end
	
	
	def export_meshes(meshbuilder, out)

		mesh_dir = meshbuilder.mesh_dir

		meshbuilder.all_meshes_to_export.each do |imesh|
			export_mesh(imesh, meshbuilder, mesh_dir, out)
		end
	end

	# Returns a Set of [Sketchup::Material, generate_uvs] pairs that are referenced by any instance in the scene.
	def get_referenced_materials(meshbuilder)
		referenced_materials = Set.new
		
		meshbuilder.instances.each do |instance|
			imesh = meshbuilder.get_indigomesh(instance)
			next if imesh.nil?

			imesh.su_materials.each do |su_mat|
				if su_mat # If the 'face material' is non-null:
					referenced_materials.insert([su_mat, false]) # add 'face' material
				else
					referenced_materials.insert([instance.material, true]) if instance.material # add 'instance' material if non-null
				end
			end
		end
		
		return referenced_materials
	end
	
	
	####################################Export Instances
	def export_instances(meshbuilder, out)
		#puts "export_instances()"
		
		meshbuilder.instances.each do |instance|

			imesh = meshbuilder.get_indigomesh(instance)

			#puts "Exporting instance, got imesh #{imesh}"
						
			# Work around issue #572 - Error when exporting scene.
			# For some reason imesh is nil here.  Need to work out why.
			if imesh.nil?
				next
			end

			trans = instance.transformation
			#definition = imesh.definition
			su_instance = instance.su_instance
							
			if su_instance and su_instance.has_indigo_keyframes?
				keyframes = su_instance.get_indigo_keyframes
				export_model(imesh, instance, trans, out, keyframes)
			else
				export_model(imesh, instance, trans, out, [])
			end
		end
	end



	######################################### Model Exporter
	# imesh - object of type IndigoMesh. (See IndigoMesh.rb)
	# instance - object of type IndigoInstance. (See IndigoMeshBuilder.rb)
	def export_model(imesh, instance, trans, out, keyframes)

		#puts "=====================export_model()=============="
		#puts "imesh: #{imesh.inspect}"
		#puts "instance: #{instance.inspect}"
		#puts "instance.su_instance: #{instance.su_instance.inspect}"
		#puts "instance.su_instance.material: #{instance.su_instance.material.inspect}" if instance.su_instance
		
		model_uid = get_and_incr_next_uid()
		
		#TEMP: new for scattering
		@last_model_uid_for_defn[instance.definition] = model_uid
		
		out.puts "	<model2>"
		out.puts "		<uid>#{model_uid}</uid>"
		write_name_string(out, 2, "name", instance.name)

		if keyframes.length > 0
			write_keyframes(trans, keyframes,out)
		else
			trans_ar = trans.to_a

			# If non-translation part of transformation is not the identity matrix, write it out:
			if (trans_ar[0] != 1.0 || trans_ar[4] != 0.0 || trans_ar[8]  != 0.0 ||
				trans_ar[1] != 0.0 || trans_ar[5] != 1.0 || trans_ar[9]  != 0.0 ||
				trans_ar[2] != 0.0 || trans_ar[6] != 0.0 || trans_ar[10] != 1.0)
			
				# printf seems to be the fastest way of writing out floating point values to a text file.
				printf(out, "		<rotation><matrix>%e %e %e %e %e %e %e %e %e</matrix></rotation>\n",
					trans_ar[0], trans_ar[4], trans_ar[8], trans_ar[1], trans_ar[5], trans_ar[9], trans_ar[2], trans_ar[6], trans_ar[10]
				)
			end
				
			out.puts "		<pos>#{trans_ar[12] * @scale} #{trans_ar[13] * @scale} #{trans_ar[14] * @scale}</pos>"
		end

		
		out.puts "		<geometry_uid>#{imesh.uid}</geometry_uid>"

		
		#  Export material UID list
		out.puts "\t\t<materials>"
		
		#puts "imesh.su_materials.size: #{imesh.su_materials.size}"
		#if imesh.su_materials.empty?
		#	puts "instance.material: #{instance.material.inspect}"
		#	if instance.su_instance.material #instance.material
		#		use_material = instance.su_instance.material # instance.material
		#		generate_uvs = true
		#		mat_uid = use_material.nil? ? @default_mat_uid : @material_uids[[use_material, generate_uvs]]
		#		
		#		puts "mat_uid: #{mat_uid}"
		#	end
		#end
		
		imesh.su_materials.each do |su_mat|
			#puts "Exporting material_uid for su_mat, material=#{su_mat.inspect}, instance.material=#{instance.material.inspect}"
			
			# Work out the material to use.  If we have a material for a face for the mesh, use that.
			# If we don't have that, use the 'instance material' instead, if it exists.
			# If it doesn't exist then just use the global default material.
			use_face_mat = su_mat != nil
			use_material = use_face_mat ? su_mat : instance.material
			generate_uvs = !use_face_mat
			mat_uid = use_material.nil? ? @default_mat_uid : @material_uids[[use_material, generate_uvs]]
			
			if !mat_uid
				puts "!!!!!!!!!!ERROR: mat_uid is nil"
				puts "use_material: #{use_material}"
				puts "generate_uvs: #{generate_uvs}"
			end
			out.puts "\t\t\t<material_uid>#{mat_uid}</material_uid>"
		end
		out.puts "\t\t</materials>"
		
		
		# Export emission scales
		imesh.su_materials.each do |su_mat|
			
			use_face_mat = su_mat != nil
			use_material = use_face_mat ? su_mat : instance.material
			generate_uvs = !use_face_mat
			
			if IndigoMaterial.emitter?(use_material)
			
				mat_uid = use_material.nil? ? @default_mat_uid : @material_uids[[use_material, generate_uvs]]
				
				# Note: a bit slow to create a new IndigoMaterial object here.
				indigo_mat = IndigoMaterial.new(use_material)
			
				export_emission_scale(indigo_mat, mat_uid, out)
			end
		end
		
		out.puts "	</model2>"
		out.puts ""
	end
	

	####
	def export_emission_scale(indigo_mat, material_uid, out)

		@light_source=true
		out.puts "		<emission_scale>"
		out.puts "			<material_uid>#{material_uid}</material_uid>"
		out.puts "			<measure>#{indigo_mat.emission_measure}</measure>" 
		out.puts "			<value>#{indigo_mat.emission_scale}</value>"
		out.puts "		</emission_scale>"
		
		if indigo_mat.ies?  # Check if material is using an IES profile.
			ies_data = indigo_mat.ies_profile  #returns in IESProfile object
			if ies_data and ies_data.valid?
				out.puts "		<ies_profile>"
				out.puts "			<material_uid>#{material_uid}</material_uid>"
				out.puts "			<path>#{File.join(self.tex_path,ies_data.filename)}</path>"
				out.puts "		</ies_profile>"
				@ies.push(ies_data.path) if (not @ies.index(ies_data.path))    # Store the ies path for future export
			end
		end
	end
	
	
	###############################Extract and write the transformation of the model
	def write_transformation(trans,out)

		trans=trans.to_a
		s = @scale

		# printf seems to be the fastest way of writing out floating point values to a text file.
		printf(out, "			<pos>%e %e %e</pos>\n			<rotation><matrix>%e %e %e %e %e %e %e %e %e</matrix></rotation>\n",
			trans[12]*s, trans[13]*s, trans[14]*s,
			trans[0], trans[4], trans[8], trans[1], trans[5], trans[9], trans[2], trans[6], trans[10]
		)
	end
		
		
		
	########
	def write_keyframe(time, trans, out)

		origin = trans.origin
		axis_angle = SkIndigo.get_axis_angle(trans)
		axis = axis_angle[0]
		angle = axis_angle[1]
		out.puts "		<keyframe>"
		out.puts "			<time>#{time.to_f}</time>"
		out.puts "			<pos>#{origin.x.to_f*@scale} #{origin.y.to_f*@scale} #{origin.z.to_f*@scale}</pos>"
		out.puts "			<rotation_quaternion>"
		out.puts "				<axis>#{axis.x} #{axis.y} #{axis.z}</axis>"
		out.puts "				<angle>#{angle}</angle>"
		out.puts "			</rotation_quaternion>"
		out.puts "		</keyframe>"
	end

	######
	def write_keyframes(trans,keyframes,out)

		out.print "			<scale>1.0</scale>\n"	

		#export the first transformation keyframe
		keyframes.each {|kf|
			time = kf[0]  #get the time
			trans = Geom::Transformation.new(kf[1])  #get the transformation
			write_keyframe(time, trans,out)
		}
	end

	######
	def write_sketchy_physics_keyframes(trans,keyframes,out)

		out.print "			<scale>1.0</scale>\n"	

		num_keyframes = trans.length.to_f  #get the number of keyframes
		#export the first transformation keyframe
		write_keyframe(0.0, trans.first, out)
		intermediate_transformations = trans[1..-2] #we have already checked prior to calling this method that there are at least two transformations in the array, so this will never return nil
		unless intermediate_transformations.empty? 
			timestep = 1.0/(num_keyframes-1)
			time = 0.0
			intermediate_transformations.each_index {|i|
				tr = intermediate_transformations[i]
				time = time + timestep
				write_keyframe(time, tr, out)
			}
		end
		write_keyframe(1.0, trans.last, out)
	end
	
	
	def export_scatter(scatter, meshbuilder, out)
		#puts "============export_scatter()=============="
		
		scatter.groups.each do |group|
			if !group.prototype_def_name
				puts "scatter.prototype_def_name was nil"
				return
			end
			
			defn = Sketchup.active_model.definitions[group.prototype_def_name]	
			if !defn
				puts "could not find defn for prototype_def_name"
				return
			end
		end
		if !scatter.target_def_name
			puts "scatter.target_def_name was nil"
			return
		end
		

		#puts "scatter.prototype_def_name: #{scatter.prototype_def_name}"
		
		
		
		#puts "scatter.target_def_name: #{scatter.target_def_name}"
		target_defn = Sketchup.active_model.definitions[scatter.target_def_name]	
		if !target_defn
			puts "could not find defn for target_def_name"
			return
		end
		
		#prototype_object_uid = @last_model_uid_for_defn[defn]
		#if !prototype_object_uid
		#	puts "could not find object UID for defn #{defn.name}"
		#	next
		#end
		
		#puts "target_defn: #{target_defn}"
		target_object_uid = @last_model_uid_for_defn[target_defn]
		if !target_object_uid
			puts "could not find UID for target_defn #{target_defn.name}"
			return
		end
		
		# geometry_uid
		# materials / material_uid
		
		#proto_instance = defn.instances[0]
		
		#puts "proto_instance: #{proto_instance.inspect}"
		
		
		
		group_geometry_uids = []
		group_mat_uids = []
		group_transforms = []
		
		scatter.groups.each do |group|
			
			defn = Sketchup.active_model.definitions[group.prototype_def_name]

			# Get a scaling and rotation matrix from the first instance of this definition, as well as any applied material.
			use_transformation = nil
			comp_instance_material = nil
			if !defn.instances.empty?
				use_transformation = defn.instances[0].transformation
				comp_instance_material = defn.instances[0].material
			else
				use_transformation = Geom::Transformation.new
				comp_instance_material = nil
			end
			
			group_transforms << use_transformation
			
			
			# Apply a transformation to the component geometry that shifts it a bit so the center of the def AABB lies over the origin.
			defn_bounds = defn.bounds
			bot_center = defn_bounds.center - Geom::Vector3d.new(0, 0, defn_bounds.depth * 0.5) # Point on the centre of the bottom of the bounding box.  Depth is z in Sketchup (8) for some reason.
			
			#puts "defn_bounds: #{defn_bounds.min}, #{defn_bounds.max}"
			#puts "defn_bounds.width: #{defn_bounds.width}"
			#puts "defn_bounds.height: #{defn_bounds.height}"
			#puts "defn_bounds.depth: #{defn_bounds.depth}"
			#puts "bot_center: #{bot_center}"
			
			translation_transform = Geom::Transformation.translation(Geom::Vector3d.new(-bot_center.x, -bot_center.y, -bot_center.z))
			#puts "getting linearised mesh for definition '#{defn.name}', guid=#{defn.guid}, inspect=#{defn.inspect}..."
			imesh = meshbuilder.get_linearised_mesh_for_definition(defn, translation_transform)
			#puts "done."
			
			# Export the mesh
			geometry_uid = export_mesh(imesh, meshbuilder, meshbuilder.mesh_dir(), out)
			group_geometry_uids << geometry_uid
			
			
			# Export any materials from the definition that aren't exported yet.
			#puts "Exporting mats for imesh #{imesh.name}..."
			#imesh.su_materials.each do |su_mat|
			#	if su_mat && su_mat.class == Sketchup::Material
			#		#puts "Exporting mat #{su_mat.name}..."
			#		export_material(IndigoMaterial.new(su_mat), out, false) # TEMP HACK Not sure about generate_uvs
			#	end
			#end
			
			
			# Compute material UIDs
			mat_uids = imesh.su_materials.map do |su_mat|
				#puts "Exporting material_uid for su_mat, material=#{su_mat.inspect}"#, instance.material=#{instance.material.inspect}"
				
				mat_uid = nil
				if su_mat.class == Sketchup::Image
					mat_uid, mat_name = export_material_for_image(su_mat, out)
				else
 					# Work out the material to use.  If we have a material for a face for the mesh, use that.
					# If we don't have that, use the 'instance material' instead, if it exists.
					# If it doesn't exist then just use the global default material. 
					use_face_mat = su_mat != nil
					use_material = use_face_mat ? su_mat : comp_instance_material
					generate_uvs = !use_face_mat
					mat_uid = use_material.nil? ? @default_mat_uid : @material_uids[[use_material, generate_uvs]]
										
					if !mat_uid && use_material
						# The material was not exported yet, we need to export it!
						export_material(IndigoMaterial.new(use_material), out, generate_uvs)
						mat_uid = @material_uids[[use_material, generate_uvs]] # Once it has been exported its UID should be in @material_uids.
					end
					
					if !mat_uid
						puts "!!!!!!!!!!ERROR: mat_uid is nil"
						puts "use_material: #{use_material}, name=#{use_material.name}"
						puts "generate_uvs: #{generate_uvs}"
						return
					end
				end
				
				mat_uid
			end
			
			group_mat_uids << mat_uids
		end
		
		
		
		out.puts "	<scatter>"
		out.puts "		<uid>#{get_and_incr_next_uid()}</uid>"
		write_name_string(out, 2, "name", scatter.name)
		
		out.puts "		<proto_object_groups>"
		index = 0
		scatter.groups.each do |group|
			out.puts "				<proto_object_group>"
			out.puts "					<geometry_uid>#{group_geometry_uids[index]}</geometry_uid>"
			
			#  Export material UID list
			out.puts "					<materials>"
			group_mat_uids[index].each { |uid| out.puts "						<material_uid>#{uid}</material_uid>" }
			out.puts "					</materials>"
			
			out.puts "					<weight>#{group.weight}</weight>"
			
			trans = group_transforms[index].to_a
			printf(out, "					<transform><matrix>%e %e %e %e %e %e %e %e %e</matrix></transform>\n",
				trans[0], trans[4], trans[8], trans[1], trans[5], trans[9], trans[2], trans[6], trans[10]
			)
			
			out.puts "				</proto_object_group>"
			index += 1
		end
		out.puts "		</proto_object_groups>"
		
		out.puts "		<target_object_uid>#{target_object_uid}</target_object_uid>"
		out.puts "		<density>#{scatter.density}</density>"
		out.puts "		<seed>#{scatter.seed}</seed>"
		out.puts "		<view_culling>#{scatter.view_culling}</view_culling>"
		out.puts "		<orient_normal>#{scatter.orient_normal}</orient_normal>"
		out.puts "		<scale_std_dev>#{scatter.scale_std_dev}</scale_std_dev>"
		out.puts "		<height_std_dev>#{scatter.height_std_dev}</height_std_dev>"
		out.puts "		<evenness>#{scatter.evenness}</evenness>"
		out.puts "	</scatter>"
		out.puts ""
	end
	
	def export_scatters(meshbuilder, out)
		#puts "export_scatters()"
		
		@scatter_settings.scatters.each do |scatter|
			export_scatter(scatter, meshbuilder, out)
		end
	end

end  #End IndigoExporter class

